# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import datetime
import getpass
import itertools
import json
import logging
import os
import platform
import re
import sys
import urllib.parse
from dataclasses import dataclass
from datetime import timedelta
from typing import Optional, Dict, List, Set

from cryptography.hazmat.primitives.asymmetric import ec, rsa
from google.protobuf.json_format import MessageToDict

from .breachwatch import BreachWatchScanCommand
from . import aliases, commands, enterprise_commands, msp_commands, msp, base
from .base import raise_parse_exception, suppress_exit, user_choice, Command, GroupCommand, as_boolean
from .helpers.record import get_record_uids as get_ruids
from .helpers.timeout import (
    enforce_timeout_range, format_timeout, get_delta_from_timeout_setting, get_timeout_setting_from_delta, parse_timeout
)
from .helpers.whoami import get_hostname, get_environment, get_data_center
from .ksm import KSMCommand, ksm_parser
from .. import __version__, vault
from .. import api, rest_api, loginv3, crypto, utils, constants, error, vault_extensions
from ..breachwatch import BreachWatch
from ..display import bcolors
from ..error import CommandError
from ..generator import KeeperPasswordGenerator, DicewarePasswordGenerator, CryptoPassphraseGenerator
from ..params import KeeperParams, LAST_RECORD_UID, LAST_FOLDER_UID, LAST_SHARED_FOLDER_UID
from ..proto import ssocloud_pb2, enterprise_pb2, APIRequest_pb2
from ..security_audit import needs_security_audit, update_security_audit_data
from ..utils import password_score
from ..vault import KeeperRecord
from ..versioning import is_binary_app, is_up_to_date_version

BREACHWATCH_MAX = 5
KEEPER_API_BATCH_LIMIT = 999  # Maximum objects per pre_delete API request

is_windows = sys.platform.startswith('win')


@dataclass
class DeletionStats:
    """Statistics for deletion operations."""
    deleted: int = 0
    failed: int = 0
    vault_changed: bool = False
    
    def add(self, other):
        """Add another DeletionStats to this one."""
        self.deleted += other.deleted
        self.failed += other.failed
        self.vault_changed = self.vault_changed or other.vault_changed


class DeletionValidator:
    """Centralized validation for deletion operations."""
    
    @staticmethod
    def validate_params(params):
        """Validate KeeperParams object."""
        if not params or not hasattr(params, 'record_cache'):
            raise ValueError("Invalid parameters provided")
    
    @staticmethod
    def validate_record_uid(record_uid):
        """Validate record UID."""
        if not record_uid:
            raise ValueError("Invalid record UID")
    
    @staticmethod
    def validate_folder(folder):
        """Validate folder object."""
        if not folder or not hasattr(folder, 'type'):
            raise ValueError("Invalid folder")
        return True
    
    @staticmethod
    def validate_api_response(rs, expected_fields=None):
        """Validate API response structure."""
        if not isinstance(rs, dict) or 'result' not in rs:
            raise ValueError("Invalid API response")
        if expected_fields:
            for field in expected_fields:
                if field not in rs:
                    raise ValueError(f"Missing field: {field}")
        return True


class FolderDeletionHelper:
    """Helper class for folder deletion operations."""
    
    @staticmethod
    def build_record_delete_object(folder, record_uid):
        """Build API delete object for a record in a folder."""
        del_obj = {
            'delete_resolution': 'unlink',
            'object_uid': record_uid,
            'object_type': 'record'
        }
        
        FolderDeletionHelper._set_folder_context(del_obj, folder)
        return del_obj
    
    @staticmethod
    def build_folder_delete_object(folder, params):
        """Build API delete object for a folder."""
        del_obj = {
            'delete_resolution': 'unlink',
            'object_uid': folder.uid,
            'object_type': folder.type
        }
        
        FolderDeletionHelper._set_parent_folder_context(del_obj, folder, params)
        return del_obj
    
    @staticmethod
    def _set_folder_context(del_obj, folder):
        """Set folder context for deletion object."""
        from ..subfolder import BaseFolderNode
        
        is_root_or_user = folder.type in {
            BaseFolderNode.RootFolderType, 
            BaseFolderNode.UserFolderType
        }
        
        if is_root_or_user:
            del_obj['from_type'] = 'user_folder'
            if folder.type == BaseFolderNode.UserFolderType:
                FolderDeletionHelper._set_folder_uid_if_present(del_obj, folder)
        else:
            del_obj['from_type'] = 'shared_folder_folder'
            FolderDeletionHelper._set_folder_uid_if_present(del_obj, folder)
    
    @staticmethod
    def _set_parent_folder_context(del_obj, folder, params):
        """Set parent folder context for folder deletion."""
        from ..subfolder import BaseFolderNode
        
        if hasattr(folder, 'parent_uid') and folder.parent_uid:
            parent = params.folder_cache.get(folder.parent_uid)
            if parent and hasattr(parent, 'uid') and hasattr(parent, 'type'):
                del_obj['from_uid'] = parent.uid
                # Shared folder containers need special handling
                if parent.type == BaseFolderNode.SharedFolderType:
                    del_obj['from_type'] = 'shared_folder_folder'
                else:
                    del_obj['from_type'] = parent.type
            else:
                # Parent not found, assume user folder
                del_obj['from_type'] = 'user_folder'
        else:
            # No parent, it's a top-level user folder
            del_obj['from_type'] = 'user_folder'
    
    @staticmethod
    def _set_folder_uid_if_present(del_obj, folder):
        """Set folder UID if present and valid."""
        if hasattr(folder, 'uid') and folder.uid:
            del_obj['from_uid'] = folder.uid

def register_commands(commands):
    commands['sync-down'] = SyncDownCommand()
    commands['this-device'] = ThisDeviceCommand()
    commands['delete-all'] = RecordDeleteAllCommand()
    commands['whoami'] = WhoamiCommand()
    commands['proxy'] = ProxyCommand()
    commands['login'] = LoginCommand()
    commands['logout'] = LogoutCommand()
    commands['check-enforcements'] = CheckEnforcementsCommand()
    commands['accept-transfer'] = AcceptTransferCommand()
    commands['delete-corrupted'] = DeleteCorruptedCommand()
    commands['echo'] = EchoCommand()
    commands['set'] = SetCommand()
    commands['help'] = HelpCommand()
    commands['secrets-manager'] = KSMCommand()
    commands['version'] = VersionCommand()
    commands['keep-alive'] = KeepAliveCommand()
    commands['generate'] = GenerateCommand()
    commands['reset-password'] = ResetPasswordCommand()
    commands['sync-security-data'] = SyncSecurityDataCommand()
    commands['blank-records'] = BlankRecordCommand()
    commands['run-as'] = RunAsCommand()
    commands['login-status'] = LoginStatusCommand()


def register_command_info(aliases, command_info):
    aliases['d'] = 'sync-down'
    aliases['delete_all'] = 'delete-all'
    aliases['gen'] = 'generate'
    aliases['v'] = 'version'
    aliases['sm'] = 'secrets-manager'
    aliases['secrets'] = 'secrets-manager'
    aliases['ssd'] = 'sync-security-data'
    for p in [sync_down_parser, whoami_parser, this_device_parser, proxy_parser, login_parser, logout_parser, echo_parser, set_parser, help_parser,
              version_parser, ksm_parser, keepalive_parser, generate_parser, reset_password_parser,
              sync_security_data_parser, delete_all_parser, loginstatus_parser, run_as_parser]:
        command_info[p.prog] = p.description


sync_down_parser = argparse.ArgumentParser(prog='sync-down', description='Download and decrypt vault data')
sync_down_parser.add_argument('-f', '--force', dest='force', action='store_true', help='full data sync')

whoami_parser = argparse.ArgumentParser(prog='whoami', description='Display information about the current user')
whoami_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='verbose output')
whoami_parser.add_argument('--json', dest='json_output', action='store_true', help='output in JSON format')
whoami_parser.error = raise_parse_exception
whoami_parser.exit = suppress_exit


this_device_available_command_verbs = ['rename', 'register', 'persistent-login', 'ip-auto-approve', 'no-yubikey-pin', 'timeout']
this_device_parser = argparse.ArgumentParser(prog='this-device', description='Display and modify settings of the current device')
this_device_parser.add_argument('ops', nargs='*', help="operation str: " + ", ".join(this_device_available_command_verbs))
this_device_parser.error = raise_parse_exception
this_device_parser.exit = suppress_exit


proxy_parser = argparse.ArgumentParser(prog='proxy', description='Sets proxy server')
proxy_parser.add_argument('-a', '--action', dest='action', action='store', choices=['list', 'add', 'remove'], help='action')
proxy_parser.add_argument('address', nargs='?', type=str, metavar="schema://[user:password@]host:port",
                          help='"add": proxy address. Schemas are "socks5h", "http", "socks4", etc')
proxy_parser.error = raise_parse_exception
proxy_parser.exit = suppress_exit


login_parser = argparse.ArgumentParser(prog='login', description='Start a login session on Commander')
login_parser.add_argument('-p', '--pass', dest='password', action='store', help='master password')
login_parser.add_argument('email', nargs='?', type=str, help='account email')
login_parser.error = raise_parse_exception
login_parser.exit = suppress_exit


logout_parser = argparse.ArgumentParser(prog='logout', description='Logout from Keeper')
logout_parser.error = raise_parse_exception
logout_parser.exit = suppress_exit


check_enforcements_parser = argparse.ArgumentParser(prog='check-enforcements',
                                                    description='Check enterprise enforcements')
check_enforcements_parser.error = raise_parse_exception
check_enforcements_parser.exit = suppress_exit


accept_transfer_parser = argparse.ArgumentParser(prog='accept-transfer', description='Accept account transfer')
accept_transfer_parser.error = raise_parse_exception
accept_transfer_parser.exit = suppress_exit


echo_parser = argparse.ArgumentParser(prog='echo', description='Displays an argument to output.')
echo_parser.add_argument('argument', nargs='?', action='store', type=str, help='argument')
echo_parser.error = raise_parse_exception
echo_parser.exit = suppress_exit


set_parser = argparse.ArgumentParser(prog='set', description='Set an environment variable.')
set_parser.add_argument('name', action='store', type=str, help='name')
set_parser.add_argument('value', action='store', type=str, help='value')
set_parser.error = raise_parse_exception
set_parser.exit = suppress_exit


help_parser = argparse.ArgumentParser(prog='help', description='Displays help on a specific command')
help_help = 'Commander\'s command (Optional -- if not specified, list of available commands is displayed)'
help_parser.add_argument('command', action='store', type=str, nargs='*',  help=help_help)
help_parser.add_argument('--legacy', dest='legacy', action='store_true', help='Show legacy/deprecated commands')
help_parser.error = raise_parse_exception
help_parser.exit = suppress_exit


version_parser = argparse.ArgumentParser(prog='version', description='Displays version of the installed Commander')
version_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='verbose output')
version_parser.add_argument('-p', '--packages', action='store_true', help='Show installed Python packages')
version_parser.error = raise_parse_exception
version_parser.exit = suppress_exit

keepalive_parser = argparse.ArgumentParser(prog='keep-alive', description='Tell the server we are here, forestalling a timeout')
keepalive_parser.error = raise_parse_exception
keepalive_parser.exit = suppress_exit

generate_parser = argparse.ArgumentParser(prog='generate', description='Generate a new password')
generate_parser.add_argument('--clipboard', '-cc', dest='clipboard', action='store_true', help='Copy to clipboard')
generate_parser.add_argument('--quiet', '-q', dest='quiet', action='store_true', help='Only print password list')
generate_parser.add_argument(
    '--password-list', '-p', dest='password_list', action='store_true',
    help='Also print password list apart from formatted table or json'
)
generate_parser.add_argument('--output', '-o', dest='output_file', action='store', help='Output to specified file')
generate_parser.add_argument(
    '--format', '-f', dest='output_format', action='store', choices=['table', 'json'],
    default='table', help='Output format for displaying password, strength, and BreachWatch if available'
)
generate_parser.add_argument(
    '--json-indent', '-i', dest='json_indent', action='store', default=2, type=int,
    help='JSON format indent (0 for compact, >0 for pretty print)'
)
generate_parser.add_argument(
    '--no-breachwatch', '-nb', dest='no_breachwatch', action='store_true',
    help='Skip BreachWatch detection if BreachWatch is enabled for this account'
)
generate_parser.add_argument(
    '--number', '-n', type=int, dest='number', action='store', help='Number of passwords', default=1
)

random_group = generate_parser.add_argument_group('Random')
random_group.add_argument(
    '--count', '-c', type=int, dest='length', action='store', help='Length of password', default=20
)
random_group.add_argument(
    '-r', '--rules', dest='rules', action='store',
    help='Use comma separated complexity integers (uppercase, lowercase, numbers, symbols)'
)
random_group.add_argument(
    '--symbols', '-s', type=int, dest='symbols', action='store',
    help='Minimum number of symbols in password or 0 for none'
)
random_group.add_argument(
    '--digits', '-d', type=int, dest='digits', action='store',
    help='Minimum number of digits in password or 0 for none'
)
random_group.add_argument(
    '--uppercase', '-u', type=int, dest='uppercase', action='store',
    help='Minimum number of uppercase letters in password or 0 for none'
)
random_group.add_argument(
    '--lowercase', '-l', type=int, dest='lowercase', action='store',
    help='Minimum number of lowercase letters in password or 0 for none'
)

passphrase_group = generate_parser.add_argument_group('Keeper Passphrase')
passphrase_group.add_argument('--recoveryphrase', dest='recoveryphrase', action='store_true',
                              help='Generate Generate a 24-word recovery phrase')

dice_group = generate_parser.add_argument_group('Diceware')
dice_group.add_argument('--dice-rolls', '-dr', type=int, dest='dice_rolls', action='store',
                        help='Number of dice rolls')
dice_group.add_argument('--delimiter', '-dl', dest='delimiter', choices=('-', '+', ':', '.', '/', '_', '='),
                        default=' ', action='store', help='Optional. Word delimiter (space if omitted)')
dice_group.add_argument('--word-list',  dest='word_list', action='store',
                        help='Optional. File path to word list')

crypto_group = generate_parser.add_argument_group('Crypto')
crypto_group.add_argument('--crypto', dest='crypto', action='store_true', help='Generate crypto wallet passphrase')

reset_password_parser = argparse.ArgumentParser(prog='reset-password', description='Reset Master Password')
reset_password_parser.add_argument('--delete-sso', dest='delete_alternate', action='store_true',
                                   help='deletes SSO master password')
reset_password_parser.add_argument('--current', '-c', dest='current_password', action='store', help='current password')
reset_password_parser.add_argument('--new', '-n', dest='new_password', action='store', help='new password')

sync_security_data_parser = argparse.ArgumentParser(prog='sync-security-data', description='Sync security audit data')
record_name_help = 'Path or UID of record whose security data is to be updated. Multiple values allowed. ' \
                   'Set to "@all" to update security data for all records.'
sync_security_data_parser.add_argument('record', type=str, action='store', nargs="+", help=record_name_help)
sync_security_data_parser.add_argument('--force', '-f', action='store_true', help='force update of security data (ignore existing security data timestamp)')
sync_security_data_parser.add_argument('--quiet', '-q', action='store_true', help='run command w/ minimal output')
sync_security_data_parser.error = raise_parse_exception
sync_security_data_parser.exit = suppress_exit


delete_all_parser = argparse.ArgumentParser(prog='delete-all', description='Delete all records and folders from the vault')
delete_all_parser.add_argument('--force', '-f', dest='force', action='store_true', 
                               help='Force deletion without confirmation prompt')
delete_all_parser.error = raise_parse_exception
delete_all_parser.exit = suppress_exit

loginstatus_parser = argparse.ArgumentParser(prog='login-status', description='Check the user login status')
loginstatus_parser.error = raise_parse_exception
loginstatus_parser.exit = suppress_exit

run_as_parser = argparse.ArgumentParser(
    prog='run-as', description='Runs application with user credentials stored on a record')
run_as_parser.add_argument('--record', '-r', dest='record', action='store', required=True,
                           help='Record name or UID')
run_as_parser.add_argument('application', help="Application to run")
run_as_parser.error = raise_parse_exception
run_as_parser.exit = suppress_exit


class SyncDownCommand(Command):
    def get_parser(self):
        return sync_down_parser

    def execute(self, params, **kwargs):
        force = kwargs.get('force') is True
        if force:
            params.revision = 0
            params.sync_down_token = b''
            if params.config:
                if 'skip_records' in params.config:
                    del params.config['skip_records']

        api.sync_down(params, record_types=force)
        if force:
            from ..loginv3 import LoginV3Flow
            LoginV3Flow.populateAccountSummary(params)

        accepted = False
        if len(params.pending_share_requests) > 0:
            for user in params.pending_share_requests:
                accepted = False
                print('Note: You have pending share request from ' + user)
                answer = user_choice('Do you want to accept these request?', 'yn', 'n')
                rq = {
                    'command': 'accept_share' if answer == 'y' else 'cancel_share',
                    'from_email': user
                }
                try:
                    rs = api.communicate(params, rq)
                    if rs['result'] == 'success':
                        accepted = accepted or answer == 'y'
                except Exception as e:
                    logging.debug('Accept share exception: %s', e)

            params.pending_share_requests.clear()

            if accepted:
                params.sync_data = True


class ThisDeviceCommand(Command):

    def get_parser(self):
        return this_device_parser

    def execute(self, params, **kwargs):

        ops = kwargs.get('ops')
        if len(ops) == 0:
            ThisDeviceCommand.print_device_info(params)
            return

        if len(ops) >= 1 and ops[0].lower() != 'register':
            if len(ops) == 1 and ops[0].lower() != 'register':
                logging.error("Must supply action and value. Available sub-commands: " + ", ".join(this_device_available_command_verbs))
                return

            if len(ops) != 2:
                logging.error("Must supply action and value. Available sub-commands: " + ", ".join(this_device_available_command_verbs))
                return

        action = ops[0].lower()

        def register_device():
            is_device_registered = loginv3.LoginV3API.register_encrypted_data_key_for_device(params)

            if is_device_registered:
                print(bcolors.OKGREEN + "Successfully registered device" + bcolors.ENDC)
            else:
                print(bcolors.OKGREEN + "Device already registered" + bcolors.ENDC)

        if action == 'rename' or action == 'ren':
            value = ops[1]
            loginv3.LoginV3API.rename_device(params, value)
            print(bcolors.OKGREEN + "Successfully renamed device to '" + value + "'" + bcolors.ENDC)

        elif action == 'register':
            register_device()

        elif action == 'persistent_login' or action == 'persistent-login' or action == 'pl':
            if ThisDeviceCommand.is_persistent_login_disabled(params):
                logging.warning('"Stay Logged In" feature is restricted by Keeper Administrator')
                return

            value = ops[1]

            value_extracted = ThisDeviceCommand.get_setting_str_to_value('persistent_login', value)
            loginv3.LoginV3API.set_user_setting(params, 'persistent_login', value_extracted)
            msg = (bcolors.OKGREEN + "ENABLED" + bcolors.ENDC) if value_extracted == '1' else (bcolors.FAIL + "DISABLED" + bcolors.ENDC)
            print("Successfully " + msg + " Persistent Login on this account")

            register_device()

            if value_extracted == '1':

                _, this_device = ThisDeviceCommand.get_account_summary_and_this_device(params)

                if this_device:
                    if 'encryptedDataKeyPresent' not in this_device:
                        print(bcolors.WARNING + "\tThis device is not registered. To register, run command `this-device register`" + bcolors.ENDC)

        elif action == 'ip_auto_approve' or action == 'ip-auto-approve' or action == 'iaa':
            value = ops[1]

            value_extracted = ThisDeviceCommand.get_setting_str_to_value('ip_disable_auto_approve', value)
            msg = (bcolors.OKGREEN + "ENABLED" + bcolors.ENDC) if value_extracted == '1' else (bcolors.FAIL + "DISABLED" + bcolors.ENDC)
            # invert ip_auto_approve value before passing it to ip_disable_auto_approve
            value_extracted = '0' if value_extracted == '1' else '1' if value_extracted == '0' else value_extracted
            loginv3.LoginV3API.set_user_setting(params, 'ip_disable_auto_approve', value_extracted)
            print("Successfully " + msg + " 'ip_auto_approve'")

        elif action == 'no-yubikey-pin':
            value = ops[1]
            value_extracted = ThisDeviceCommand.get_setting_str_to_value('no-yubikey-pin', value)
            msg = (bcolors.OKGREEN + "ENABLED" + bcolors.ENDC) if value_extracted == '0' else (bcolors.FAIL + "DISABLED" + bcolors.ENDC)
            loginv3.LoginV3API.set_user_setting(params, 'security_keys_no_user_verify', value_extracted)
            print("Successfully " + msg + " Security Key PIN verification")

        elif action == 'timeout' or action == 'to':

            value = ops[1]
            timeout_delta = enforce_timeout_range(ThisDeviceCommand.get_setting_str_to_value('logout_timer', value))
            loginv3.LoginV3API.set_user_setting(params, 'logout_timer', get_timeout_setting_from_delta(timeout_delta))
            dispay_value = 'default value' if timeout_delta == timedelta(0) else format_timeout(timeout_delta)
            print(f'Successfully set "logout_timer" to {dispay_value}.')

        else:
            raise Exception("Unknown sub-command " + action + ". Available sub-commands: ", ", ".join(this_device_available_command_verbs))

    @staticmethod
    def is_persistent_login_disabled(params):  # type: (KeeperParams) -> bool
        if params.enforcements and 'booleans' in params.enforcements:
            return next((x['value'] for x in params.enforcements['booleans'] if x['key'] == 'restrict_persistent_login'), False)
        else:
            return False

    @staticmethod
    def get_setting_str_to_value(name: str, value: str):

        name = name.lower()
        value = value.lower()

        if name == 'persistent_login' or name == 'ip_disable_auto_approve' or name == 'no-yubikey-pin':
            final_val = '1' if as_boolean(value) else '0'
        elif name == 'logout_timer':
            final_val = parse_timeout(value)
        else:
            raise Exception("Unhandled settings name '" + name + "'")

        return final_val

    @staticmethod
    def get_account_summary_and_this_device(params: KeeperParams):
        def to_alphanumerics(text):
            # remove ALL non - alphanumerics
            return re.sub(r'[\W_]+', '', text)

        def compare_device_tokens(t1: str, t2: str):
            t1 = to_alphanumerics(t1)
            t2 = to_alphanumerics(t2)

            return t1 == t2

        acct_summary = loginv3.LoginV3API.accountSummary(params)
        acct_summary_dict = MessageToDict(acct_summary)

        devices = acct_summary_dict['devices']
        current_device_token = params.config['device_token']

        this_device = next((item for item in devices if compare_device_tokens(item['encryptedDeviceToken'], current_device_token)), None)

        return acct_summary_dict, this_device

    @staticmethod
    def print_device_info(params: KeeperParams):
        acct_summary_dict, this_device = ThisDeviceCommand.get_account_summary_and_this_device(params)

        print('{:>32}: {}'.format('Device Name', this_device['deviceName']))
        # print("{:>32}: {}".format('API Client Version', rest_api.CLIENT_VERSION))

        if 'encryptedDataKeyPresent' in this_device:
            print("{:>32}: {}".format('Data Key Present', (bcolors.OKGREEN + 'YES' + bcolors.ENDC) if this_device['encryptedDataKeyPresent'] else (bcolors.FAIL + 'NO' + bcolors.ENDC)))
        else:
            print("{:>32}: {}".format('Data Key Present', (bcolors.FAIL + 'missing' + bcolors.ENDC)))

        if 'ipDisableAutoApprove' in acct_summary_dict['settings']:
            ipDisableAutoApprove = acct_summary_dict['settings']['ipDisableAutoApprove']
            # ip_disable_auto_approve - If enabled, the device is NOT automatically approved
            # If disabled, the device will be auto approved
            ipAutoApprove = not ipDisableAutoApprove
            print("{:>32}: {}".format('IP Auto Approve',
                                      (bcolors.OKGREEN + 'ON' + bcolors.ENDC)
                                      if ipAutoApprove else
                                      (bcolors.FAIL + 'OFF' + bcolors.ENDC)))
        else:
            print("{:>32}: {}".format('IP Auto Approve', (bcolors.OKGREEN + 'ON' + bcolors.ENDC)))
            # ip_disable_auto_approve = 0 / disabled (default) <==> IP Auto Approve :ON

        persistentLogin = acct_summary_dict['settings'].get('persistentLogin', False)
        print("{:>32}: {}".format('Persistent Login',
                                  (bcolors.OKGREEN + 'ON' + bcolors.ENDC)
                                  if persistentLogin and not ThisDeviceCommand.is_persistent_login_disabled(params) else
                                  (bcolors.FAIL + 'OFF' + bcolors.ENDC)))

        no_user_verify = acct_summary_dict['settings'].get('securityKeysNoUserVerify', False)
        print("{:>32}: {}".format(
            'Security Key No PIN', (bcolors.OKGREEN + 'ON' + bcolors.ENDC)
            if no_user_verify else (bcolors.FAIL + 'OFF' + bcolors.ENDC)))

        if 'securityKeysNoUserVerify' in acct_summary_dict['settings']:
            device_timeout = get_delta_from_timeout_setting(acct_summary_dict['settings']['logoutTimer'])
            print("{:>32}: {}".format('Device Logout Timeout', format_timeout(device_timeout)))

        if 'logoutTimer' in acct_summary_dict['settings']:
            device_timeout = get_delta_from_timeout_setting(acct_summary_dict['settings']['logoutTimer'])
            print("{:>32}: {}".format('Device Logout Timeout', format_timeout(device_timeout)))

        else:
            device_timeout = timedelta(hours=1)
            print("{:>32}: Default".format('Logout Timeout'))

        if 'Enforcements' in acct_summary_dict and 'longs' in acct_summary_dict['Enforcements']:
            logout_timeout = next((x['value'] for x in acct_summary_dict['Enforcements']['longs']
                                    if x['key'] == 'logout_timer_desktop'), None)
            if logout_timeout:
                enterprise_timeout = timedelta(minutes=int(logout_timeout))
                print("{:>32}: {}".format('Enterprise Logout Timeout', format_timeout(enterprise_timeout)))

                print("{:>32}: {}".format('Effective Logout Timeout',
                                          format_timeout(min(enterprise_timeout, device_timeout))))

        print('{:>32}: {}'.format('Is SSO User', params.settings['sso_user'] if 'sso_user' in params.settings else False))

        print('{:>32}: {}'.format('Config file', params.config_filename))

        print("\nAvailable sub-commands: ", bcolors.OKBLUE + (", ".join(this_device_available_command_verbs)) + bcolors.ENDC)


class RecordDeleteAllCommand(Command):
    """
    Delete all records and folders from the vault.

    Args:
        params: KeeperParams instance
        force: Skip confirmation if True (set internally after user confirms)
        
    Returns:
        None
        
    Note:
        Records in multiple folders will be processed using the first 
        folder found. Deletions are processed in batches of 999 items.
    """
    
    def get_parser(self):
        return delete_all_parser

    def execute(self, params, **kwargs):
        """Delete all records and folders from vault."""
        try:
            DeletionValidator.validate_params(params)
        except ValueError:
            logging.error("Invalid parameters provided")
            return
        
        if not self._confirm_user_wants_deletion(kwargs):
            return
        
        kwargs['force'] = True
        api.sync_down(params)
        
        record_stats = self._process_record_deletion(params)
        folder_stats = self._process_folder_deletion(params)
        
        self._display_summary(record_stats, folder_stats)
        self._finalize_deletion(params, record_stats, folder_stats)

    def _confirm_user_wants_deletion(self, kwargs):
        """Show warning and get user confirmation."""
        from ..display import bcolors
        
        # Check if user already passed --force flag
        force_flag_passed = kwargs.get('force', False)
        
        if force_flag_passed:
            # Force flag was passed via command line, skip confirmation
            logging.info("Force flag detected, proceeding without confirmation...")
            print(f"{bcolors.WARNING}Force mode: Deleting all records and folders without confirmation{bcolors.ENDC}")
            return True
        
        # Show confirmation prompt
        confirmation_msg = (
            f"{bcolors.FAIL}{bcolors.BOLD}WARNING: This will permanently delete ALL records and ALL folders from your vault!{bcolors.ENDC}\n"
            f"{bcolors.WARNING}This action cannot be undone.{bcolors.ENDC}\n"
            f"Are you sure you want to proceed?"
        )
        
        uc = user_choice(confirmation_msg, 'yn', default='n')
        if uc.lower() != 'y':
            logging.info("Operation cancelled by user")
            return False
        
        return True

    def _process_record_deletion(self, params):
        """Collect and delete all records."""
        records_with_folders = self._collect_records_with_folders(params)
        if not records_with_folders:
            logging.info('No records found to delete')
            return DeletionStats()
        
        logging.info('Preparing to delete %s records from Keeper', len(records_with_folders))
        return self._delete_objects_in_batches(params, records_with_folders, 'records')

    def _collect_records_with_folders(self, params):
        """Collect all records with their folder contexts."""
        from ..subfolder import find_all_folders
        
        records_with_folders = []
        
        if len(params.record_cache) == 0:
            return records_with_folders
        
        for record_uid in params.record_cache.keys():
            try:
                DeletionValidator.validate_record_uid(record_uid)
                
                # Find all folders containing this record
                folders = list(find_all_folders(params, record_uid))
                if folders:
                    folder = folders[0]  # Use first folder for API context
                else:
                    folder = params.root_folder
                
                DeletionValidator.validate_folder(folder)
                records_with_folders.append((folder, record_uid))
                
            except ValueError:
                logging.warning("Skipping record %s: invalid record UID", record_uid)
                continue
        
        return records_with_folders

    def _process_folder_deletion(self, params):
        """Delete all empty folders."""
        return self._delete_empty_folders(params)

    def _delete_objects_in_batches(self, params, objects_with_context, object_type_name):
        """Generic batch deletion handler for records."""
        stats = DeletionStats()
        
        while objects_with_context:
            current_batch = objects_with_context[:KEEPER_API_BATCH_LIMIT]
            objects_with_context = objects_with_context[KEEPER_API_BATCH_LIMIT:]
            
            batch_stats = self._delete_batch(params, current_batch, object_type_name)
            stats.add(batch_stats)
        
        return stats

    def _delete_batch(self, params, batch, object_type_name):
        """Delete a single batch of objects."""
        stats = DeletionStats()
        
        # Build pre_delete request
        rq = self._build_pre_delete_request(batch)
        
        if not rq['objects']:
            logging.info(f"No valid {object_type_name} in this batch to delete")
            return stats
        
        try:
            # Execute pre_delete
            rs = api.communicate(params, rq)
            DeletionValidator.validate_api_response(rs, ['pre_delete_response'])
            
            if rs['result'] == 'success':
                batch_stats = self._execute_deletion_batch(params, rs, len(batch))
                stats.add(batch_stats)
            else:
                stats.failed += len(batch)
                logging.warning(f"Pre-delete failed for batch of {len(batch)} {object_type_name}: {rs.get('message', 'Unknown error')}")
                
        except Exception:
            stats.failed += len(batch)
            logging.error(f"Error processing batch of {len(batch)} {object_type_name}")
        
        return stats

    def _build_pre_delete_request(self, batch):
        """Build pre_delete request from batch of objects."""
        rq = {
            'command': 'pre_delete',
            'objects': []
        }
        
        for folder, record_uid in batch:
            try:
                del_obj = FolderDeletionHelper.build_record_delete_object(folder, record_uid)
                rq['objects'].append(del_obj)
            except Exception:
                logging.warning(f"Skipping object {record_uid}: invalid object")
                continue
        
        return rq

    def _execute_deletion_batch(self, params, pre_delete_response, batch_size):
        """Execute the actual deletion after pre_delete."""
        stats = DeletionStats()
        
        pdr = pre_delete_response.get('pre_delete_response')
        if not pdr:
            stats.failed += batch_size
            logging.error("Missing pre_delete_response in API response")
            return stats
        
        # Show what would be deleted
        would_delete = pdr.get('would_delete', {})
        summary = would_delete.get('deletion_summary', [])
        for x in summary:
            print(x)
        
        # Validate pre_delete_token
        pre_delete_token = pdr.get('pre_delete_token')
        if not pre_delete_token:
            stats.failed += batch_size
            logging.error("Missing pre_delete_token in API response")
            return stats
        
        # Execute actual deletion
        delete_rq = {
            'command': 'delete',
            'pre_delete_token': pre_delete_token
        }
        
        delete_rs = api.communicate(params, delete_rq)
        
        try:
            DeletionValidator.validate_api_response(delete_rs)
            if delete_rs['result'] == 'success':
                stats.deleted += batch_size
                stats.vault_changed = True
                logging.info(f"Batch of {batch_size} objects deleted successfully")
            else:
                stats.failed += batch_size
                logging.warning(f"Failed to delete batch of {batch_size} objects: {delete_rs.get('message', 'Unknown error')}")
        except ValueError:
            stats.failed += batch_size
            logging.error("Invalid deletion response received")
        
        return stats

    def _display_summary(self, record_stats, folder_stats):
        """Display deletion summary."""
        if record_stats.deleted > 0:
            logging.info("Total records deleted: %s", record_stats.deleted)
        if record_stats.failed > 0:
            logging.warning("Total records failed to delete: %s", record_stats.failed)
        if folder_stats > 0:
            logging.info("Total folders deleted: %s", folder_stats)

    def _finalize_deletion(self, params, record_stats, folder_stats):
        """Finalize the deletion process."""
        if record_stats.vault_changed or folder_stats > 0:
            params.revision = 0
            params.sync_data = True

    def _delete_empty_folders(self, params):
        """
        Delete all empty folders after records have been deleted.
        
        Args:
            params: KeeperParams instance
            
        Returns:
            int: Number of folders successfully deleted
            
        Note:
            Only deletes folders that are completely empty (no records and no subfolders).
            Root folder is never deleted. Folders are sorted by depth (deepest first)
            to avoid parent-child dependency issues.
        """
        from ..subfolder import BaseFolderNode
        
        # Security check: Ensure params are valid
        if not params or not hasattr(params, 'folder_cache'):
            logging.error("Invalid parameters provided to _delete_empty_folders")
            return 0
        
        # Sync down to get latest folder state after record deletion
        api.sync_down(params)
        
        # Find all folders that are safe to delete
        empty_folders = []
        for folder_uid, folder in params.folder_cache.items():
            # Security check: Skip if folder or folder_uid is invalid
            if not folder_uid or not folder or not hasattr(folder, 'type'):
                logging.warning("Skipping invalid folder entry")
                continue
            
            # Security check: Do not attempt to delete root folder
            # Delete all other folder types (user folders, shared folders, shared folder subfolders)
            if folder.type == BaseFolderNode.RootFolderType:
                continue
            
            # Check if folder is empty (no records)
            records_in_folder = params.subfolder_record_cache.get(folder_uid, set())
            if len(records_in_folder) == 0:
                # Also check if it has any subfolders - only delete if completely empty
                if not folder.subfolders or len(folder.subfolders) == 0:
                    empty_folders.append(folder)
        
        if not empty_folders:
            logging.info("No folders found to delete")
            return 0
        
        logging.info('Found %s folders to delete', len(empty_folders))
        
        # Sort folders by depth (deepest first) to avoid parent-child dependency issues
        # Folders with longer paths are deeper in the hierarchy
        from ..subfolder import get_folder_path
        empty_folders.sort(key=lambda f: len(get_folder_path(params, f.uid).split('/')), reverse=True)
        
        total_folders_deleted = 0
        total_folders_failed = 0
        
        folders_to_delete = empty_folders[:]
        while len(folders_to_delete) > 0:
            # Prepare pre_delete request for folders
            rq = {
                'command': 'pre_delete',
                'objects': []
            }
            
            # Process chunk of folders
            chunk = folders_to_delete[:KEEPER_API_BATCH_LIMIT]
            folders_to_delete = folders_to_delete[KEEPER_API_BATCH_LIMIT:]
            
            for folder in chunk:
                # Security check: Validate folder has required attributes
                if not hasattr(folder, 'uid') or not hasattr(folder, 'type'):
                    logging.warning("Skipping folder without required attributes")
                    continue
                
                # Security check: Validate folder type is deletable (not root)
                if folder.type == BaseFolderNode.RootFolderType:
                    logging.warning("Skipping root folder - cannot be deleted")
                    continue
                
                del_obj = FolderDeletionHelper.build_folder_delete_object(folder, params)
                
                rq['objects'].append(del_obj)
            
            # Skip empty batches (all folders were filtered out by security checks)
            if not rq['objects']:
                logging.info("No valid folders in this batch to delete")
                continue
            
            # Send pre_delete request for folders
            try:
                rs = api.communicate(params, rq)
                # Security check: Validate API response
                if not isinstance(rs, dict) or 'result' not in rs:
                    logging.error("Invalid API response received")
                    total_folders_failed += len(chunk)
                    continue
                
                if rs['result'] == 'success':
                    pdr = rs.get('pre_delete_response')
                    if not pdr:
                        logging.error("Missing pre_delete_response in API response")
                        total_folders_failed += len(chunk)
                        continue
                    
                    # Show what would be deleted
                    would_delete = pdr.get('would_delete', {})
                    summary = would_delete.get('deletion_summary', [])
                    for x in summary:
                        print(x)
                    
                    # Execute deletion automatically (force is always True after initial confirmation in execute())
                    # Security check: Validate pre_delete_token exists
                    pre_delete_token = pdr.get('pre_delete_token')
                    if not pre_delete_token:
                        logging.error("Missing pre_delete_token in API response")
                        total_folders_failed += len(chunk)
                        continue
                    
                    # Execute actual folder deletion
                    delete_rq = {
                        'command': 'delete',
                        'pre_delete_token': pre_delete_token
                    }
                    delete_rs = api.communicate(params, delete_rq)
                    # Security check: Validate deletion response
                    if not isinstance(delete_rs, dict) or 'result' not in delete_rs:
                        logging.error("Invalid deletion response received")
                        total_folders_failed += len(chunk)
                    elif delete_rs['result'] == 'success':
                        total_folders_deleted += len(chunk)
                        logging.info("Batch of %s folders deleted successfully", len(chunk))
                    else:
                        total_folders_failed += len(chunk)
                        logging.warning("Failed to delete batch of %s folders: %s", 
                                      len(chunk), delete_rs.get('message', 'Unknown error'))
                else:
                    total_folders_failed += len(chunk)
                    logging.warning("Pre-delete failed for batch of %s folders: %s", 
                                  len(chunk), rs.get('message', 'Unknown error'))
            except Exception:
                total_folders_failed += len(chunk)
                logging.error("Error processing batch of %s folders", len(chunk))
        
        if total_folders_failed > 0:
            logging.warning("Total folders failed to delete: %s", total_folders_failed)
        
        return total_folders_deleted


class WhoamiCommand(Command):
    def get_parser(self):
        return whoami_parser

    def execute(self, params, **kwargs):
        json_output = kwargs.get('json_output', False)
        verbose = kwargs.get('verbose', False)
        
        if json_output:
            # Collect data for JSON output
            data = {}
            
            if params.session_token:
                data['logged_in'] = True
                hostname = get_hostname(params.rest_context.server_base)
                data['user'] = params.user
                data['server'] = hostname
                data['data_center'] = get_data_center(hostname)
                
                environment = get_environment(hostname)
                if environment:
                    data['environment'] = environment
                
                if params.license:
                    account_type = params.license['account_type'] if 'account_type' in params.license else None
                    if account_type == 2:
                        data['admin'] = params.enterprise is not None
                    
                    account_type_name = 'Enterprise' if account_type == 2 \
                        else 'Family Plan' if account_type == 1 \
                        else params.license['product_type_name']
                    data['account_type'] = account_type_name
                    data['renewal_date'] = params.license['expiration_date']
                    
                    if 'bytes_total' in params.license:
                        storage_bytes = int(params.license['bytes_total'])
                        storage_gb = storage_bytes >> 30
                        storage_bytes_used = params.license['bytes_used'] if 'bytes_used' in params.license else 0
                        data['storage_capacity'] = f'{storage_gb}GB'
                        storage_usage = (int(storage_bytes_used) * 100 // storage_bytes) if storage_bytes != 0 else 0
                        data['storage_usage'] = f'{storage_usage}%'
                        data['storage_renewal_date'] = params.license['storage_expiration_date']
                    
                    data['breachwatch'] = params.license.get('breach_watch_enabled', False)
                    if params.enterprise:
                        data['reporting_and_alerts'] = params.license.get('audit_and_reporting_enabled', False)

                if verbose:
                    data['records_count'] = len(params.record_cache)
                    sf_count = len(params.shared_folder_cache)
                    if sf_count > 0:
                        data['shared_folders_count'] = sf_count
                    team_count = len(params.team_cache)
                    if team_count > 0:
                        data['teams_count'] = team_count

                if params.enterprise:
                    enterprise_licenses = []
                    for x in params.enterprise.get('licenses', []):
                        license_info = {}
                        product_type_id = x.get('product_type_id', 0)
                        tier = x.get('tier', 0)
                        if product_type_id in (3, 5):
                            plan = 'Enterprise' if tier == 1 else 'Business'
                        elif product_type_id in (9, 10):
                            distributor = x.get('distributor', False)
                            plan = 'Distributor' if distributor else 'Managed MSP'
                        elif product_type_id in (11, 12):
                            plan = 'Keeper MSP'
                        elif product_type_id == 8:
                            plan = 'MC ' + 'Enterprise' if tier == 1 else 'Business'
                        else:
                            plan = 'Unknown'
                        if product_type_id in (5, 10, 12):
                            plan += ' Trial'
                        license_info['base_plan'] = plan
                        
                        paid = x.get('paid') is True
                        if paid:
                            exp = x.get('expiration')
                            if exp > 0:
                                dt = datetime.datetime.fromtimestamp(exp // 1000) + datetime.timedelta(days=1)
                                n = datetime.datetime.now()
                                td = (dt - n).days
                                expires = str(dt.date())
                                if td > 0:
                                    expires += f' (in {td} days)'
                                else:
                                    expires += ' (expired)'
                                license_info['license_expires'] = expires
                        
                        license_info['user_licenses'] = {
                            'plan': x.get("number_of_seats", ""),
                            'active': x.get("seats_allocated", ""),
                            'invited': x.get("seats_pending", "")
                        }
                        
                        file_plan = x.get('file_plan')
                        file_plan_lookup = {x[0]: x[2] for x in constants.ENTERPRISE_FILE_PLANS}
                        license_info['secure_file_storage'] = file_plan_lookup.get(file_plan, '')
                        
                        addons = []
                        addon_lookup = {a[0]: a[1] for a in constants.MSP_ADDONS}
                        for ao in x.get('add_ons'):
                            if isinstance(ao, dict):
                                enabled = ao.get('enabled') is True
                                if enabled:
                                    name = ao.get('name')
                                    addon_name = addon_lookup.get(name) or name
                                    if name == 'secrets_manager':
                                        api_count = ao.get('api_call_count')
                                        if isinstance(api_count, int) and api_count > 0:
                                            addon_name += f' ({api_count:,} API calls)'
                                    elif name == 'connection_manager':
                                        seats = ao.get('seats')
                                        if isinstance(seats, int) and seats > 0:
                                            addon_name += f' ({seats} licenses)'
                                    addons.append(addon_name)
                        if addons:
                            license_info['add_ons'] = addons
                        
                        enterprise_licenses.append(license_info)
                    
                    if enterprise_licenses:
                        data['enterprise_licenses'] = enterprise_licenses
            else:
                data['logged_in'] = False
                data['message'] = 'Not logged in'
            
            # Output JSON
            import json
            print(json.dumps(data, indent=2))
        else:
            # Original formatted output
            if params.session_token:
                hostname = get_hostname(params.rest_context.server_base)
                print('{0:>20s}: {1:<20s}'.format('User', params.user))
                print('{0:>20s}: {1:<20s}'.format('Server', hostname))
                print('{0:>20s}: {1:<20s}'.format('Data Center', get_data_center(hostname)))
                environment = get_environment(hostname)
                if environment:
                    print('{0:>20s}: {1:<20s}'.format('Environment', get_environment(hostname)))
                if params.license:
                    account_type = params.license['account_type'] if 'account_type' in params.license else None
                    if account_type == 2:
                        display_admin = 'No' if params.enterprise is None else 'Yes'
                        print('{0:>20s}: {1:<20s}'.format('Admin', display_admin))

                    print('')
                    account_type_name = 'Enterprise' if account_type == 2 \
                        else 'Family Plan' if account_type == 1 \
                        else params.license['product_type_name']
                    print('{0:>20s}: {1:<20s}'.format('Account Type', account_type_name))
                    print('{0:>20s}: {1:<20s}'.format('Renewal Date', params.license['expiration_date']))
                    if 'bytes_total' in params.license:
                        storage_bytes = int(params.license['bytes_total'])  # note: int64 in protobuf in python produces string as opposed to an int or long.
                        storage_gb = storage_bytes >> 30
                        storage_bytes_used = params.license['bytes_used'] if 'bytes_used' in params.license else 0
                        print('{0:>20s}: {1:<20s}'.format('Storage Capacity', f'{storage_gb}GB'))
                        storage_usage = (int(storage_bytes_used) * 100 // storage_bytes) if storage_bytes != 0 else 0     # note: int64 in protobuf in python produces string  as opposed to an int or long.
                        print('{0:>20s}: {1:<20s}'.format('Usage', f'{storage_usage}%'))
                        print('{0:>20s}: {1:<20s}'.format('Storage Renewal Date', params.license['storage_expiration_date']))
                    print('{0:>20s}: {1:<20s}'.format('BreachWatch', 'Yes' if params.license.get('breach_watch_enabled') else 'No'))
                    if params.enterprise:
                        print('{0:>20s}: {1:<20s}'.format('Reporting & Alerts', 'Yes' if params.license.get('audit_and_reporting_enabled') else 'No'))

                if verbose:
                    print('')
                    print('{0:>20s}: {1}'.format('Records', len(params.record_cache)))
                    sf_count = len(params.shared_folder_cache)
                    if sf_count > 0:
                        print('{0:>20s}: {1}'.format('Shared Folders', sf_count))
                    team_count = len(params.team_cache)
                    if team_count > 0:
                        print('{0:>20s}: {1}'.format('Teams', team_count))

                if params.enterprise:
                    print('')
                    print('{0:>20s}:'.format('Enterprise License'))
                    for x in params.enterprise.get('licenses', []):
                        product_type_id = x.get('product_type_id', 0)
                        tier = x.get('tier', 0)
                        if product_type_id in (3, 5):
                            plan = 'Enterprise' if tier == 1 else 'Business'
                        elif product_type_id in (9, 10):
                            distributor = x.get('distributor', False)
                            plan = 'Distributor' if distributor else 'Managed MSP'
                        elif product_type_id in (11, 12):
                            plan = 'Keeper MSP'
                        elif product_type_id == 8:
                            plan = 'MC ' + 'Enterprise' if tier == 1 else 'Business'
                        else:
                            plan = 'Unknown'
                        if product_type_id in (5, 10, 12):
                            plan += ' Trial'
                        print('{0:>20s}: {1}'.format('Base Plan', plan))
                        paid = x.get('paid') is True
                        if paid:
                            exp = x.get('expiration')
                            if exp > 0:
                                dt = datetime.datetime.fromtimestamp(exp // 1000) + datetime.timedelta(days=1)
                                n = datetime.datetime.now()
                                td = (dt - n).days
                                expires = str(dt.date())
                                if td > 0:
                                    expires += f' (in {td} days)'
                                else:
                                    expires += ' (expired)'
                                print('{0:>20s}: {1}'.format('Expires', expires))
                        print('{0:>20s}: {1}'.format('User Licenses', f'Plan: {x.get("number_of_seats", "")}    Active: {x.get("seats_allocated", "")}    Invited: {x.get("seats_pending", "")}'))
                        file_plan = x.get('file_plan')
                        file_plan_lookup = {x[0]: x[2] for x in constants.ENTERPRISE_FILE_PLANS}
                        print('{0:>20s}: {1}'.format('Secure File Storage', file_plan_lookup.get(file_plan, '')))
                        addons = []
                        addon_lookup = {a[0]: a[1] for a in constants.MSP_ADDONS}
                        for ao in x.get('add_ons'):
                            if isinstance(ao, dict):
                                enabled = ao.get('enabled') is True
                                if enabled:
                                    name = ao.get('name')
                                    addon_name = addon_lookup.get(name) or name
                                    if name == 'secrets_manager':
                                        api_count = ao.get('api_call_count')
                                        if isinstance(api_count, int) and api_count > 0:
                                            addon_name += f' ({api_count:,} API calls)'
                                    elif name == 'connection_manager':
                                        seats = ao.get('seats')
                                        if isinstance(seats, int) and seats > 0:
                                            addon_name += f' ({seats} licenses)'
                                    addons.append(addon_name)
                        for i, addon in enumerate(addons):
                            print('{0:>20s}: {1}'.format('Secure Add Ons' if i == 0 else '', addon))
            else:
                print('{0:>20s}:'.format('Not logged in'))
    

class VersionCommand(Command):
    def get_parser(self):
        return version_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):
        version_details = is_up_to_date_version(params)
        is_verbose = kwargs.get('verbose', False)
        show_packages = kwargs.get('packages', False)

        this_app_version = __version__

        if version_details.get('is_up_to_date') is None:
            this_app_version = f'{this_app_version} (Current version)'

        if not is_verbose:
            print('{0}: {1}'.format('Commander Version', this_app_version))
        else:
            print('{0:>20s}: {1}'.format('Commander Version', this_app_version))
            print("{0:>20s}: {1}".format('API Client Version', rest_api.CLIENT_VERSION))
            print('{0:>20s}: {1}'.format('Python Version', sys.version.replace("\n", "")))
            print('{0:>20s}: {1}'.format('Operating System', loginv3.CommonHelperMethods.get_os() + '(' + platform.release() + ')'))
            print('{0:>20s}: {1}'.format('Working directory', os.getcwd()))
            print('{0:>20s}: {1}'.format('Package directory', os.path.dirname(api.__file__)))
            print('{0:>20s}: {1}'.format('Config. File', params.config_filename))
            print('{0:>20s}: {1}'.format('Executable', sys.executable))

        if logging.getLogger().isEnabledFor(logging.DEBUG) or show_packages:
            ver = sys.version_info
            if ver.major >= 3 and ver.minor >= 8:
                import importlib.metadata
                dist = importlib.metadata.packages_distributions()
                packages = {}
                for pack in dist.values():
                    if isinstance(pack, list) and len(pack) > 0:
                        name = pack[0]
                        if name in packages:
                            continue
                        try:
                            version = importlib.metadata.version(name)
                            packages[name] = version
                        except Exception as e:
                            logging.debug('Get package %s version error: %s', name, e)
                installed_packages_list = [f'{x[0]}=={x[1]}' for x in packages.items()]
                installed_packages_list.sort(key=lambda x: x.lower())
                print('{0:>20s}: {1}'.format('Packages', installed_packages_list))

        if version_details.get('is_up_to_date') is None:
            logging.debug("It appears that Commander is up to date")
        elif not version_details.get('is_up_to_date'):

            latest_version = version_details.get('current_github_version')

            print((bcolors.WARNING +
                   'Latest Commander Version: %s\n'
                   'You can download the current version at: %s \n' + bcolors.ENDC)
                  % (latest_version, version_details.get('new_version_download_url')))
        if is_binary_app():
            print("Installation path: {0} ".format(sys._MEIPASS))


class KeepAliveCommand(Command):
    """Just issue a keepalive to keep the interactive session from timing out."""
    def get_parser(self):
        """Return the argparse parser.  This one has no options, but we want a help message anyway."""
        return keepalive_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        """Just send the keepalive."""
        api.send_keepalive(params)

class ProxyCommand(Command):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return proxy_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):  # type: (KeeperParams, any) -> any
        action = kwargs.get('action')
        if action == 'add':
            proxy_server = kwargs.get('address')  # type: str
            if proxy_server:
                is_valid = False
                for prefix in {'socks5', 'http', 'https'}:
                    if proxy_server.startswith(prefix):
                        is_valid = True
                        break
                if not is_valid:
                    logging.warning('Proxy server "%s" does not appear to be a valid proxy URL', proxy_server)
                    proxy_server = ''
            else:
                logging.warning('"add" action requires "proxy" parameter.')

            if proxy_server:
                params.proxy = proxy_server
            else:
                return
        elif action == 'remove':
            params.proxy = None

        if params.proxy:
            logging.info('Proxy server: %s', params.proxy)
        else:
            logging.info('Proxy is not configured.')


class LoginCommand(Command):
    def get_parser(self):
        return login_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):
        if msp.current_mc_id:
            msp.current_mc_id = None
            msp.mc_params_dict.clear()

        new_login = kwargs.get('new_login') is True
        if new_login:
            params.clear_session()

        user = kwargs.get('email') or ''
        password = kwargs.get('password') or ''

        try:
            if not user:
                user = input('... {0:>16}: '.format('User(Email)')).strip()
            if not user:
                return
        except KeyboardInterrupt as e:
            logging.info('Canceled')
            return

        params.user = user.lower()
        if not password and isinstance(params.config, dict):
            if 'user' in params.config and 'password' in params.config:
                if params.config['user'] == params.user:
                    password = params.config['password']

        params.password = password

        try:
            api.login(params, new_login=new_login)
        except Exception as exc:
            logging.warning(str(exc))

        if params.session_token:
            SyncDownCommand().execute(params, force=True)
            if params.is_enterprise_admin:
                api.query_enterprise(params, True)
            try:
                if params.breach_watch:
                    BreachWatchScanCommand().execute(params, suppress_no_op=True)
                if params.enterprise_ec_key:
                    SyncSecurityDataCommand().execute(params, record='@all', suppress_no_op=True)
            except Exception as e:
                logging.warning(f'A problem was encountered while updating BreachWatch/security data: {e}')
                logging.debug(e, exc_info=True)


class CheckEnforcementsCommand(Command):
    def get_parser(self):
        return check_enforcements_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):
        if params.enforcements:
            if 'enterprise_invited' in params.enforcements:
                print('You\'ve been invited to join {0}.'.format(params.enforcements['enterprise_invited']))
                action = user_choice('A(ccept)/D(ecline)/I(gnore)?: ', 'adi')
                action = action.lower()
                if action == 'a':
                    action = 'accept'
                elif action == 'd':
                    action = 'decline'
                if action in ['accept', 'decline']:
                    e_rq = {
                        'command': '{0}_enterprise_invite'.format(action)
                    }
                    if action == 'accept':
                        verification_code = input('Please enter the verification code sent via email: ')
                        if verification_code:
                            e_rq['verification_code'] = verification_code
                        else:
                            e_rq = None
                    if e_rq:
                        try:
                            api.communicate(params, e_rq)
                            logging.info('%s enterprise invite', 'Accepted' if action == 'accept' else 'Declined')
                            #TODO reload enterprise settings
                        except Exception as e:
                            logging.error('Enterprise %s failure: %s', action, e)

        share_account_by = params.get_share_account_timestamp()
        if share_account_by is not None:
            account_transfer_command = AcceptTransferCommand()
            account_transfer_command.execute(params, **kwargs)


class AcceptTransferCommand(Command):
    def get_parser(self):
        return check_enforcements_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):
        share_account_by = params.get_share_account_timestamp()
        if share_account_by is not None:
            if api.accept_account_transfer_consent(params):
                if 'must_perform_account_share_by' in params.settings:
                    del params.settings['must_perform_account_share_by']
                if 'share_account_to' in params.settings:
                    del params.settings['share_account_to']
                logging.info('Account transfer accepted.')
            else:
                logging.info('Account transfer canceled.')
        else:
            logging.info('There is no account transfer to accept.')


class LogoutCommand(Command):
    def get_parser(self):
        return logout_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):
        if msp.current_mc_id:
            msp.current_mc_id = None
            msp.mc_params_dict.clear()

        if params.session_token:
            try:
                api.communicate_rest(params, None, 'vault/logout_v3')
            except:
                pass

        # Clean up Rust WebRTC tube registry if it exists
        try:
            from .tunnel.port_forward.tunnel_helpers import cleanup_tube_registry
            cleanup_tube_registry(params)
        except Exception as e:
            logging.debug('Tube registry cleanup error: %s', e)

        if params.sso_login_info and 'idp_session_id' in params.sso_login_info:
            sso_url = params.sso_login_info.get('sso_url') or ''
            sp_url_builder = urllib.parse.urlparse(sso_url)
            sp_url_query = urllib.parse.parse_qsl(sp_url_builder.query)
            session_id = params.sso_login_info.get('idp_session_id') or ''
            if params.sso_login_info.get('is_cloud'):
                sso_rq = ssocloud_pb2.SsoCloudRequest()
                sso_rq.clientVersion = rest_api.CLIENT_VERSION
                sso_rq.embedded = True
                sso_rq.username = params.user.lower()
                sso_rq.idpSessionId = session_id
                transmission_key = utils.generate_aes_key()
                rq_payload = APIRequest_pb2.ApiRequestPayload()
                rq_payload.apiVersion = 3
                rq_payload.payload = sso_rq.SerializeToString()
                api_rq = APIRequest_pb2.ApiRequest()
                api_rq.locale = params.rest_context.locale or 'en_US'

                server_public_key = rest_api.SERVER_PUBLIC_KEYS[params.rest_context.server_key_id]
                if isinstance(server_public_key, rsa.RSAPublicKey):
                    api_rq.encryptedTransmissionKey = crypto.encrypt_rsa(transmission_key, server_public_key)
                elif isinstance(server_public_key, ec.EllipticCurvePublicKey):
                    api_rq.encryptedTransmissionKey = crypto.encrypt_ec(transmission_key, server_public_key)
                else:
                    raise ValueError('Invalid server public key')
                api_rq.publicKeyId = params.rest_context.server_key_id
                api_rq.encryptedPayload = crypto.encrypt_aes_v2(rq_payload.SerializeToString(), transmission_key)
                sp_url_query.append(('payload', utils.base64_url_encode(api_rq.SerializeToString())))
            else:
                sp_url_query.append(('embedded', ''))
                sp_url_query.append(('token', ''))
                sp_url_query.append(('user', params.user.lower()))
                if session_id:
                    sp_url_query.append(('session_id', session_id))

            sp_url_builder = sp_url_builder._replace(path=sp_url_builder.path.replace('/login', '/logout'), query=urllib.parse.urlencode(sp_url_query, doseq=True))
            sp_url = urllib.parse.urlunparse(sp_url_builder)
            logging.info('SSO Logout URL\n%s', sp_url)

        params.clear_session()


class EchoCommand(Command):
    def get_parser(self):
        return echo_parser

    def execute(self, params, **kwargs):
        argument = kwargs.get('argument')
        if argument:
            print(argument)
        else:
            envs = {LAST_RECORD_UID, LAST_FOLDER_UID, LAST_SHARED_FOLDER_UID}
            for name in params.environment_variables:
                envs.add(name)
            names = [x for x in envs]
            names.sort()
            for name in names:
                if name in params.environment_variables:
                    print('${{{0}}} = "{1}"'.format(name, params.environment_variables[name] ))
                else:
                    print('${{{0}}} ='.format(name))


class SetCommand(Command):
    def get_parser(self):
        return set_parser

    def execute(self, params, **kwargs):
        name = kwargs['name']
        value = kwargs.get('value')
        if value:
            params.environment_variables[name] = value
        else:
            if name in params.environment_variables:
                del params.environment_variables[name]


class HelpCommand(Command):
    def get_parser(self):
        return help_parser

    def execute(self, params, **kwargs):
        help_commands = kwargs.get('command')
        show_legacy = kwargs.get('legacy', False)
        if not help_commands:
            from ..cli import display_command_help
            display_command_help(params.enterprise_ec_key, show_legacy=show_legacy)
            return

        if isinstance(help_commands, list) and len(help_commands) > 0:
            cmd = help_commands[0]
            help_commands = help_commands[1:]
            if cmd in aliases:
                ali = aliases[cmd]
                if type(ali) == tuple:
                    cmd = ali[0]
                else:
                    cmd = ali

            if cmd in commands:
                command = commands[cmd]
            elif cmd in enterprise_commands:
                command = enterprise_commands[cmd]
            elif cmd in msp_commands:
                command = msp_commands[cmd]
            else:
                command = None

            if isinstance(command, Command):
                parser = command.get_parser()
                if parser:
                    parser.print_help()
            elif isinstance(command, GroupCommand):
                if len(help_commands) == 0:
                    command.print_help(command=cmd)
                else:
                    while len(help_commands) > 0:
                        cmd = help_commands[0]
                        help_commands = help_commands[1:]
                        if cmd in command.subcommands:
                            subcommand = command.subcommands[cmd]
                            if isinstance(subcommand, Command):
                                parser = subcommand.get_parser()
                                if parser:
                                    parser.print_help()
                                break
                            elif isinstance(subcommand, GroupCommand):
                                command = subcommand
                    command.print_help(command=cmd)

    def is_authorised(self):
        return False


class DeleteCorruptedCommand(Command):
    def execute(self, params, **kwargs):
        bad_records = set()
        for record_uid in params.record_cache:
            record = params.record_cache[record_uid]
            if not record.get('data_unencrypted'):
                if record_uid in params.record_owner_cache:
                    own = params.record_owner_cache[record_uid]
                    if own.owner is True:
                        bad_records.add(record_uid)
        if len(bad_records) > 0:
            uc = user_choice('Do you want to delete {0} corrupted records?'.format(len(bad_records)), 'yn', default='n')
            if uc.lower() == 'y':
                request = {
                    'command': 'record_update',
                    'delete_records': list(bad_records)
                }
                logging.info('Deleting %s records from Keeper', len(params.record_cache))
                response_json = api.communicate(params, request)
                success = [info for info in response_json['delete_records'] if info['status'] == 'success']
                if len(success) > 0:
                    logging.info("%s records deleted successfully", len(success))
                failures = [info for info in response_json['delete_records'] if info['status'] != 'success']
                if len(failures) > 0:
                    logging.warning("%s records failed to delete", len(failures))
        else:
            logging.info('No corrupted records are found.')


class GenerateCommand(Command):
    def get_parser(self):
        return generate_parser

    def execute(self, params, number=None, no_breachwatch=None,
                length=None, symbols=None, digits=None, uppercase=None, lowercase=None, rules=None,
                output_format=None, output_file=None, json_indent=None, quiet=False, password_list=False,
                clipboard=False, return_result=False, **kwargs):
        """
        Executes "generate" command

        Parameters
        ----------
        params : KeeperParams
            Connected KeeperParams instance
        number : int
            Number of passwords. Default: 1
        length : int
            Length of password. Default: 20
        symbols : int
            Minimum number of symbols in password if positive exact if 0 or negative. Default: None
        digits : int
            Minimum number of digits in password if positive exact if 0 or negative. Default: None
        uppercase : int
            Minimum number of uppercase letters in password if positive exact if 0 or negative. Default: None
        lowercase : int
            Minimum number of lowercase letters in password if positive exact if 0 or negative. Default: None
        rules : str
            Use comma separated complexity integers (uppercase, lowercase, numbers, symbols)
        no_breachwatch : bool
            Skip BreachWatch detection if BreachWatch is enabled for this account
        output_format : str
            Output format for displaying password, strength, and BreachWatch if available. 'table' or 'json'
        output_file : str
            File name to store result. stdout is omitted
        json_indent : int
            JSON format indent (0 for compact, >0 for pretty print). Default: 2
        quiet : bool
            Only print password list
        password_list: bool
            Also print password list apart from formatted table or json
        clipboard: bool
            Copy to clipboard
        return_result : bool
            If True return tuple of password dict and formatted output string
        """

        if kwargs.get('crypto') is True:
            kpg = CryptoPassphraseGenerator()
        if kwargs.get('recoveryphrase') is True:
            kpg = DicewarePasswordGenerator(24, word_list_file='bip-39.english.txt', delimiter=' ')
        elif isinstance(kwargs.get('dice_rolls'), int):
            dice_rolls = kwargs.get('dice_rolls')
            delimiter = kwargs.get('delimiter') or ' '
            kpg = DicewarePasswordGenerator(dice_rolls, word_list_file=kwargs.get('word_list'), delimiter=delimiter)
        else:
            if rules and all(i is None for i in (symbols, digits, uppercase, lowercase)):
                kpg = KeeperPasswordGenerator.create_from_rules(rules, length)
                if kpg is None:
                    logging.warning('Using default password complexity rules')
                    kpg = KeeperPasswordGenerator(length=length)
            else:
                if rules:
                    logging.warning(
                        'Ignoring "rules" option used with "symbols", "digits", "uppercase", or "lowercase" option'
                    )
                kpg = KeeperPasswordGenerator(
                    length=length, symbols=symbols, digits=digits, caps=uppercase, lower=lowercase
                )

        get_new_password_count = number
        no_breachwatch = no_breachwatch or getattr(params, 'breach_watch', None) is None

        passwords = []    # type: List[Dict]
        breachwatch_count = 0
        while len(passwords) < get_new_password_count:
            new_passwords = [kpg.generate() for i in range(get_new_password_count - len(passwords))]
            if no_breachwatch:
                passwords = [{'password': p, 'strength': password_score(p)} for p in new_passwords]

            else:
                euids = []
                breachwatch_count += 1
                breachwatch_maxed = breachwatch_count >= BREACHWATCH_MAX
                for breach_result in params.breach_watch.scan_passwords(params, new_passwords):
                    pw = breach_result[0]
                    if breach_result[1].euid:
                        euids.append(breach_result[1].euid)
                    if breach_result[1].breachDetected:
                        if breachwatch_maxed:
                            passwords.append(
                                {'password': pw, 'strength': password_score(pw), 'breach_watch': 'Failed'}
                            )
                    else:
                        passwords.append(
                            {'password': pw, 'strength': password_score(pw), 'breach_watch': 'Passed'}
                        )
                params.breach_watch.delete_euids(params, euids)

        if quiet:
            formatted_output = ''
        elif output_format == 'table':
            breach_watch = '' if no_breachwatch else '{breach_watch:13}'
            format_template = '{count:<5}{strength:<13}' + breach_watch + '{password}'
            header = format_template.format(
                count='', strength='Strength(%)', breach_watch='BreachWatch', password='Password'
            )
            password_output = [format_template.format(count=i, **p) for i, p in enumerate(passwords, start=1)]
            formatted_output = header + '\n' + '\n'.join(password_output)
        elif output_format == 'json':
            formatted_output = json.dumps(passwords, indent=json_indent or None)
        else:
            formatted_output = ''

        if quiet or password_list:
            skip_line = '\n\n' if password_list else ''
            formatted_output += skip_line + '\n'.join(p['password'] for p in passwords)

        if clipboard:
            import pyperclip
            pyperclip.copy(formatted_output)
            logging.info('New passwords copied to clipboard')
        elif not output_file:
            print(formatted_output)

        if output_file:
            try:
                with open(output_file, 'w', encoding='utf-8') as f:
                    f.write(formatted_output)
            except Exception as e:
                logging.warning('Error writing to file {}: {}'.format(output_file, str(e)))
            else:
                logging.info('Wrote to file {}'.format(output_file))

        if return_result:
            return passwords, formatted_output


class ResetPasswordCommand(Command):
    def get_parser(self):
        return reset_password_parser

    def execute(self, params, **kwargs):
        current_password = kwargs.get('current_password')
        is_sso_user = params.settings.get('sso_user', False)
        if is_sso_user:
            allow_alternate_passwords = False
            if 'booleans' in params.enforcements:
                allow_alternate_passwords = next((x.get('value') for x in params.enforcements['booleans']
                                                  if x.get('key') == 'allow_alternate_passwords'), False)
            if not allow_alternate_passwords:
                logging.warning('You do not have the required privilege to perform this operation.')
                return

        try:
            current_salt = api.communicate_rest(params, None, 'authentication/get_salt_and_iterations',
                                                rs_type=APIRequest_pb2.Salt)
        except error.KeeperApiError as kae:
            if is_sso_user and kae.result_code == 'doesnt_exist':
                current_salt = None
            else:
                raise kae

        is_delete_alternate = kwargs.get('delete_alternate')
        if is_delete_alternate:
            if is_sso_user:
                logging.info('Deleting SSO Master Password for \"%s\"', params.user)
            else:
                logging.warning('\"%s\" in not SSO account', params.user)
                return
        else:
            if is_sso_user:
                logging.info('%s SSO Master Password for \"%s\"',
                             'Changing' if current_salt else 'Setting', params.user)
            else:
                logging.info('Changing Master Password for \"%s\"', params.user)

        if current_salt:
            auth_hash = crypto.derive_keyhash_v1(current_password, current_salt.salt, current_salt.iterations)
            rq = APIRequest_pb2.MasterPasswordReentryRequest()
            rq.pbkdf2Password = utils.base64_url_encode(auth_hash)
            rq.action = APIRequest_pb2.UNMASK
            try:
                rs = api.communicate_rest(params, rq, 'authentication/validate_master_password',
                                          rs_type=APIRequest_pb2.MasterPasswordReentryResponse, payload_version=1)
                if rs.status != APIRequest_pb2.MP_SUCCESS:
                    logging.info('Failed to change password')
            except:
                logging.warning('Current password incorrect')
                return
        else:
            current_password = ''

        if is_delete_alternate:
            if current_salt:
                uid_rq = APIRequest_pb2.UidRequest()
                uid_rq.uid.append(current_salt.uid)
                api.communicate_rest(params, uid_rq, 'authentication/delete_v2_alternate_password')
                logging.info('SSO Master Password has been deleted')
            else:
                logging.info('SSO Master password is not found')
            return

        new_password = kwargs.get('new_password')
        if not new_password:
            password1 = getpass.getpass(prompt='{0:>24}: '.format('New Password'), stream=None).strip()
            password2 = getpass.getpass(prompt='{0:>24}: '.format('Re-enter New Password'), stream=None).strip()
            print('')
            if password1 != password2:
                logging.warning('New password does not match')
                return
            if current_password and password1 == current_password:
                logging.warning('Please choose a different password')
                return
            new_password = password1

        rules_rq = enterprise_pb2.DomainPasswordRulesRequest()
        rules_rq.username = params.user
        rules_rs = api.communicate_rest(params, rules_rq, 'authentication/get_domain_password_rules',
                                        rs_type=APIRequest_pb2.NewUserMinimumParams)
        failed_rules = []
        for i in range(len(rules_rs.passwordMatchRegex)):
            rule = rules_rs.passwordMatchRegex[i]
            is_match = re.match(rule, new_password)
            if not is_match:
                failed_rules.append(rules_rs.passwordMatchDescription[i])
        if failed_rules:
            logging.warning('Password rules:\n%s', '\n'.join((f'  {x}' for x in failed_rules)))
            return

        if params.breach_watch:
            euids = []
            for result in params.breach_watch.scan_passwords(params, [new_password]):
                if result[1].euid:
                    euids.append(result[1].euid)
                logging.info('Breachwatch password scan result: %s', 'WEAK' if result[1].breachDetected else 'GOOD')
            if euids:
                params.breach_watch.delete_euids(params, euids)
        else:
            score = utils.password_score(new_password)
            logging.info('Password strength: %s', 'WEAK' if score < 40 else 'FAIR' if score < 60 else 'MEDIUM' if score < 80 else 'STRONG')

        iterations = current_salt.iterations if current_salt else constants.PBKDF2_ITERATIONS
        iterations = max(iterations, constants.PBKDF2_ITERATIONS)

        auth_salt = crypto.get_random_bytes(16)
        if is_sso_user:
            ap_rq = APIRequest_pb2.UserAuthRequest()
            ap_rq.uid = current_salt.uid if current_salt else crypto.get_random_bytes(16)
            ap_rq.salt = auth_salt
            ap_rq.iterations = iterations
            ap_rq.authHash = crypto.derive_keyhash_v1(new_password, auth_salt, iterations)
            key = crypto.derive_keyhash_v2('data_key', new_password, auth_salt, iterations)
            ap_rq.encryptedDataKey = crypto.encrypt_aes_v2(params.data_key, key)
            ap_rq.encryptedClientKey = crypto.encrypt_aes_v2(params.client_key, key)
            ap_rq.loginType = APIRequest_pb2.ALTERNATE
            ap_rq.name = current_salt.name if current_salt else 'alternate'
            api.communicate_rest(params, ap_rq, 'authentication/set_v2_alternate_password')
            logging.info(f'SSO Master Password has been {("changed" if current_salt else "set")}')
        else:
            auth_verifier = utils.create_auth_verifier(new_password, auth_salt, iterations)
            data_salt = crypto.get_random_bytes(16)
            encryption_params = utils.create_encryption_params(new_password, data_salt, iterations, params.data_key)
            mp_rq = {
                'command': 'change_master_password',
                'auth_verifier': utils.base64_url_encode(auth_verifier),
                'encryption_params': utils.base64_url_encode(encryption_params)
            }
            api.communicate(params, mp_rq)
            logging.info('Master Password has been changed')


class SyncSecurityDataCommand(Command):
    def get_parser(self):
        return sync_security_data_parser

    def execute(self, params, **kwargs):
        if not params.enterprise_ec_key:
            msg = 'Command not allowed -- This command is limited to enterprise users only.'
            raise CommandError('sync-security-data', msg)

        def parse_input_records():  # type: () -> Set[str]
            names = kwargs.get('record',[])
            do_all = '@all' in names
            return params.record_cache.keys() if do_all \
                else itertools.chain.from_iterable(get_ruids(params, n) for n in names)

        force_update = kwargs.get('force', False)
        api.sync_down(params)
        recs = [KeeperRecord.load(params, r) for r in parse_input_records()]
        should_update = lambda r: force_update or bool(needs_security_audit(params, r))
        recs_to_update = [r for r in recs if should_update(r)]
        num_to_update = len(recs_to_update)
        num_updated = update_security_audit_data(params, recs_to_update)
        if num_updated:
            BreachWatch.save_reused_pw_count(params)
            api.sync_down(params)
        if not kwargs.get('quiet'):
            if num_updated:
                logging.info(f'Updated security data for [{num_updated}] record(s)')
            elif not kwargs.get('suppress_no_op') and not num_to_update:
                logging.info('No records requiring security-data updates found')


class BlankRecordCommand(Command):
    blank_record_parser = argparse.ArgumentParser(prog='blank-record', parents=[base.report_output_parser],
                                                  description='returns empy records')
    def get_parser(self):
        return BlankRecordCommand.blank_record_parser

    def execute(self, params, **kwargs):
        headers = ['record_uid', 'record_type', 'record_title']
        table = []
        for record in vault_extensions.find_records(params, record_version=(2,3)):
            if isinstance(record, vault.PasswordRecord):
                if not record.login and not record.password and not record.link:
                    table.append([record.record_uid, '', record.title])
            elif isinstance(record, vault.TypedRecord):
                all_empty = True
                for field in record.fields:
                    if field.type == 'fileRef' and record.record_type != 'file':
                        continue
                    if not field.value:
                        continue
                    if isinstance(field.value, list):
                        for value in field.value:
                            if isinstance(value, str) and not value:
                                continue
                            all_empty = False
                            break
                    if not all_empty:
                        break
                if all_empty:
                    table.append([record.record_uid, record.record_type, record.title])

        fmt = kwargs.get('format') or 'table'
        output = kwargs.get('output')
        return base.dump_report_data(table, headers, fmt=fmt, filename=output)


class RunAsCommand(Command):
    def get_parser(self):
        return run_as_parser

    def execute(self, params, **kwargs):
        from .. import native

        record_name = kwargs.get('record')
        record = base.RecordMixin.resolve_single_record(params, record_name)
        if not record:
            raise CommandError('', f'Record \"{record_name}\" not found.')

        username = base.RecordMixin.get_record_field(record, 'login')
        if not username:
            raise CommandError('', f'Username not found on record \"{record.title}\"')

        password = base.RecordMixin.get_record_field(record, 'password')
        if not password:
            raise CommandError('', f'Password not found on record \"{record.title}\"')

        try:
            native.run_as(username, password, kwargs.get('application'))
        except OSError as e:
            raise CommandError('', str(e))


class LoginStatusCommand(Command):
    def get_parser(self):
        return loginstatus_parser

    def is_authorised(self):
        return False

    def execute(self, params, **kwargs):
        if not params.user:
            print("Not logged in")
            return
        
        has_clone_code = bool(params.clone_code)
        has_session = bool(params.session_token)
        
        if not has_clone_code and not has_session:
            print("Not logged in")
            return
        
        if has_session:
            print("Logged in")
            return
        
        if has_clone_code:
            try:
                class NonInteractiveLoginUi:
                    def on_device_approval(self, step):
                        step.cancel()
                    
                    def on_two_factor(self, step):
                        step.cancel()
                    
                    def on_password(self, step):
                        step.cancel()
                    
                    def on_sso_redirect(self, step):
                        step.cancel()
                    
                    def on_sso_data_key(self, step):
                        step.cancel()
                
                api.login(params, new_login=False, login_ui=NonInteractiveLoginUi())
                
                if params.session_token:
                    print("Logged in")
                else:
                    print("Not logged in")
                    
            except Exception as e:
                logging.debug(f"Non-interactive login failed: {e}")
                print("Not logged in")
        else:
            print("Not logged in")


from typing import Set, Optional

from ... import api
from ...params import KeeperParams
from ...subfolder import try_resolve_path


# Get record UID(s) given one of its identifiers: name (if current folder contains the record), path, or UID
def get_record_uids(params, name):  # type: (KeeperParams, str) -> Set[Optional[str]]
    uids = set()
    if name in params.record_cache:
        uids = [name]
    else:
        rs = try_resolve_path(params, name, find_all_matches=True)
        # if rs is not None:
        folders, name = rs
        if folders and name is not None:
            for folder in folders:
                folder_uid = folder.uid or ''
                if folder_uid in params.subfolder_record_cache:
                    for r_uid in params.subfolder_record_cache[folder_uid]:
                        r = api.get_record(params, r_uid)
                        if r.title.lower() == name.lower():
                            uids.add(r_uid)
    return uids



from __future__ import annotations

import abc
import datetime
import json
import logging
from typing import List, Optional, Set, Iterable, Tuple, Dict, Any, cast

from ..params import KeeperParams
from . import admin_storage, admin_types
from .. import crypto, utils, api
from ..proto import pedm_pb2, folder_pb2
from ..storage import types as storage_types, in_memory
#from ...enterprise import enterprise_loader, sqlite_enterprise_storage
#from ...authentication import notifications, endpoint


class RebuildTask:
    def __init__(self, full_rebuild = False) -> None:
        self._full_rebuild = full_rebuild
        self._agents: Optional[Set[str]] = None
        self._policies: Optional[Set[str]] = None
        self._collections: Optional[Set[str]] = None
        self._approvals: Optional[Set[str]] = None

    @property
    def full_rebuild(self) -> bool:
        return self._full_rebuild

    def add_agents(self, agents: Iterable[str]) -> None:
        if self._full_rebuild:
            return
        if self._agents is None:
            self._agents = set()
        self._agents.update(agents)

    def add_policies(self, policies: Iterable[str]) -> None:
        if self._full_rebuild:
            return
        if self._policies is None:
            self._policies = set()
        self._policies.update(policies)

    def add_collections(self, collections: Iterable[str]) -> None:
        if self._full_rebuild:
            return
        if self._collections is None:
            self._collections = set()
        self._collections.update(collections)

    def add_approvals(self, approvals: Iterable[str]) -> None:
        if self._full_rebuild:
            return
        if self._approvals is None:
            self._approvals = set()
        self._approvals.update(approvals)

    @property
    def agents(self) -> Optional[Iterable[str]]:
        return self._agents

    @property
    def policies(self) -> Optional[Iterable[str]]:
        return self._policies

    @property
    def collections(self) -> Optional[Iterable[str]]:
        return self._collections

    @property
    def approvals(self) -> Optional[Iterable[str]]:
        return self._approvals


class IPedmAdmin(abc.ABC):
    @abc.abstractmethod
    def sync_down(self, *, reload: bool = False) -> None:
        pass

    @property
    @abc.abstractmethod
    def deployments(self) -> storage_types.IEntityReader[admin_types.PedmDeployment, str]:
        pass

    @property
    @abc.abstractmethod
    def agents(self) -> storage_types.IEntityReader[admin_types.PedmAgent, str]:
        pass


# class PedmAdminNotifications(notifications.BasePushNotifications):
#     def __init__(self, *, on_message: Callable[[Union[str, bytes]], None]):
#         super().__init__()
#         self.on_message = on_message
#
#     def on_messaged_received(self, message: Union[str, bytes]):
#         self.on_message(message)
#
#     async def on_connected(self):
#         pass


class PedmPlugin(IPedmAdmin):
    def __init__(self, params: KeeperParams):
        self.params = params
        if params.enterprise is None:
            raise Exception("Not an enterprise admin")
        self._enterprise_id = params.enterprise_id
        self.enterprise_uid: str = utils.base64_url_encode(self._enterprise_id.to_bytes(16, byteorder='big'))
        self.storage: admin_storage.IPedmStorage = admin_storage.SqlitePedmStorage(params.get_connection, self._enterprise_id)

        self.device_uid = utils.generate_uid()
        self._populate_data = True
        self._agent_key: Optional[bytes] = None
        self._all_agents: Optional[bytes] = None

        self._deployments = in_memory.InMemoryEntityStorage[admin_types.PedmDeployment, str]()
        self._agents = in_memory.InMemoryEntityStorage[admin_types.PedmAgent, str]()
        self._deployment_agents = in_memory.InMemoryLinkStorage[admin_types.PedmDeploymentAgent, str, str]()
        self._policies = in_memory.InMemoryEntityStorage[admin_types.PedmPolicy, str]()
        self._collections = in_memory.InMemoryEntityStorage[admin_types.PedmCollection, str]()
        self._approvals = in_memory.InMemoryEntityStorage[admin_types.PedmApproval, str]()

        # self._push_notifications = PedmAdminNotifications(on_message=self.on_push_message)
        # self.connect_pushes()
        self._need_sync = True
        self.logger = logging.getLogger("keeper.pedm")

    """
    def connect_pushes(self):
        auth = self.loader.keeper_auth


        if 'ROUTER_URL' in os.environ:
            router_url = os.environ['ROUTER_URL']
        else:
            router_url = f'https://{auth.keeper_endpoint.get_router_server()}'
        up = urlparse(router_url)
        url_comp = ('wss' if up.scheme == 'https' else 'ws', up.netloc, 'api/user/client', None, None, None)
        url = str(urlunparse(url_comp))

        transmission_key = utils.generate_aes_key()
        session_token = auth.auth_context.session_token
        encrypted_session_token = crypto.encrypt_aes_v2(session_token, transmission_key)
        encrypted_transmission_key = endpoint.encrypt_with_keeper_key(
            transmission_key, auth.keeper_endpoint.server_key_id)

        headers = {
            'TransmissionKey': base64.b64encode(encrypted_transmission_key).decode('ascii'),
            'Authorization': 'KeeperUser ' + base64.b64encode(encrypted_session_token).decode('ascii'),
        }
        asyncio.run_coroutine_threadsafe(
            self._push_notifications.main_loop(url=url, headers=headers), background.get_loop())

    def on_push_message(self, message: Union[str, bytes]):
        self.logger.debug('Received PEDM Admin notification: %s', message)
        if isinstance(message, str):
            try:
                data = json.loads(message)
                message_type = data.get('type')
                if message_type == 'pedm_sync':
                    self._need_sync = True
            except:
                pass
    """

    def close(self):
        pass

    @property
    def deployments(self) -> storage_types.IEntityReader[admin_types.PedmDeployment, str]:
        return self._deployments

    @property
    def agents(self) -> storage_types.IEntityReader[admin_types.PedmAgent, str]:
        return self._agents

    @property
    def policies(self) -> storage_types.IEntityReader[admin_types.PedmPolicy, str]:
        return self._policies

    @property
    def collections(self) -> storage_types.IEntityReader[admin_types.PedmCollection, str]:
        return self._collections

    @property
    def deployment_agents(self) -> storage_types.ILinkReader[admin_types.PedmDeploymentAgent, str, str]:
        return self._deployment_agents

    @property
    def approvals(self) -> storage_types.IEntityReader[admin_types.PedmApproval, str]:
        return self._approvals

    @property
    def need_sync(self) -> bool:
        return self._need_sync

    @property
    def agent_key(self) -> bytes:
        if self._agent_key is None:
            enterprise = self.params.enterprise
            tree_key = enterprise['unencrypted_tree_key']
            encrypted_ec_private_key = utils.base64_url_decode(enterprise['keys']['ecc_encrypted_private_key'])
            ec_private_key = crypto.decrypt_aes_v2(encrypted_ec_private_key, tree_key)

            x1 = int.from_bytes(tree_key[:16], byteorder='big', signed=False)
            x2 = int.from_bytes(tree_key[16:], byteorder='big', signed=False)
            salt = (x1 ^ x2).to_bytes(length=16, byteorder='big', signed=False)
            self._agent_key = crypto.derive_key_v1(utils.base64_url_encode(ec_private_key), salt, 1_000_000)
        return self._agent_key

    @property
    def all_agents(self) -> bytes:
        if self._all_agents is None:
            self._all_agents = (0).to_bytes(16, byteorder='big')
        return self._all_agents

    def build_data(self, task: RebuildTask) -> None:
        enterprise = self.params.enterprise
        tree_key = enterprise['unencrypted_tree_key']

        self._deployments.clear()
        deps: List[admin_types.PedmDeployment] = []
        for dep in self.storage.deployments.get_all_entities():
            try:
                pd = self.load_deployment(dep, tree_key)
                deps.append(pd)
            except Exception as e:
                self.logger.debug('Deployment "%s" decryption error: %s', dep.deployment_uid, e)
        self._deployments.put_entities(deps)

        if task.full_rebuild:
            self._agents.clear()
            self._policies.clear()
            self._collections.clear()
            self._approvals.clear()
        else:
            if task.agents is not None:
                self._agents.delete_uids(task.agents)
            if task.policies is not None:
                self._policies.delete_uids(task.policies)
            if task.collections is not None:
                self._collections.delete_uids(task.collections)
            if task.approvals is not None:
                self._approvals.delete_uids(task.approvals)

        def get_agents() -> Iterable[admin_storage.PedmStorageAgent]:
            if task.full_rebuild:
                yield from self.storage.agents.get_all_entities()
            if task.agents is not None:
                for agent_uid in task.agents:
                    a = self.storage.agents.get_entity(agent_uid)
                    if a is not None:
                        yield a

        ags: List[admin_types.PedmAgent] = []
        for agent_dto in get_agents():
            properties: Optional[Dict[str, Any]] = None
            if agent_dto.data is not None and len(agent_dto.data) > 0:
                try:
                    decrypted_data = crypto.decrypt_aes_v2(agent_dto.data, self.agent_key)
                    properties = json.loads(decrypted_data)
                except Exception as e:
                    self.logger.debug('Agent "%s" decryption error: %s', agent_dto.agent_uid, e)
            agent = admin_types.PedmAgent(
                agent_uid=agent_dto.agent_uid, machine_id=agent_dto.machine_id, created=agent_dto.created,
                deployment_uid=agent_dto.deployment_uid, disabled=agent_dto.disabled, public_key=agent_dto.public_key,
                properties=properties)
            ags.append(agent)
        self._agents.put_entities(ags)

        self._deployment_agents.clear()
        das: List[admin_types.PedmDeploymentAgent] = []
        for agent in self._agents.get_all_entities():
            if agent.deployment_uid:
                das.append(admin_types.PedmDeploymentAgent(deployment_uid=agent.deployment_uid, agent_uid=agent.agent_uid))
        self._deployment_agents.put_links(das)

        def get_policies() -> Iterable[admin_storage.PedmStoragePolicy]:
            if task.full_rebuild:
                yield from self.storage.policies.get_all_entities()
            elif task.policies is not None:
                for policy_uid in task.policies:
                    p = self.storage.policies.get_entity(policy_uid)
                    if p is not None:
                        yield p

        policies: List[admin_types.PedmPolicy] = []
        for policy_dto in get_policies():
            try:
                policy_key = crypto.decrypt_aes_v2(policy_dto.key, self.agent_key)
                json_data = crypto.decrypt_aes_v2(policy_dto.data, policy_key)
                data = json.loads(json_data)
                admin_data = json.loads(policy_dto.admin_data)
                policy = admin_types.PedmPolicy(policy_uid=policy_dto.policy_uid, policy_key=policy_key, data=data,
                                                admin_data=admin_data, disabled=policy_dto.disabled)
                policies.append(policy)
            except Exception as e:
                self.logger.debug('Policy load error: %s', e)
        if len(policies) > 0:
            self._policies.put_entities(policies)

        def get_collections() -> Iterable[admin_storage.PedmStorageCollection]:
            if task.full_rebuild:
                yield from self.storage.collections.get_all_entities()
            elif task.collections is not None:
                for collection_uid in task.collections:
                    c = self.storage.collections.get_entity(collection_uid)
                    if c is not None:
                        yield c

        collections: List[admin_types.PedmCollection] = []
        for collection_dto in get_collections():
            try:
                collection_value = crypto.decrypt_aes_v2(collection_dto.data, self.agent_key).decode('utf-8')
                collection_data = json.loads(collection_value)
                collection = admin_types.PedmCollection(
                    collection_uid=collection_dto.collection_uid, collection_type=collection_dto.collection_type,
                    collection_data=collection_data, created=collection_dto.created)
            except Exception as e:
                self.logger.info('Collection "%s" load error: %s', collection_dto.collection_uid, e)
                collection = admin_types.PedmCollection(
                    collection_uid=collection_dto.collection_uid, collection_type=collection_dto.collection_type,
                    collection_data={}, created=collection_dto.created)
            collections.append(collection)

        if len(collections) > 0:
            self._collections.put_entities(collections)

        def get_approvals() -> Iterable[admin_storage.PedmStorageApproval]:
            if task.full_rebuild:
                yield from self.storage.approvals.get_all_entities()
            elif task.approvals is not None:
                for approval_uid in task.approvals:
                    c = self.storage.approvals.get_entity(approval_uid)
                    if c is not None:
                        yield c

        approvals: List[admin_types.PedmApproval] = []
        for approval_dto in get_approvals():
            try:
                application_info = json.loads(crypto.decrypt_aes_v2(approval_dto.application_info, self.agent_key))
                account_info = json.loads(crypto.decrypt_aes_v2(approval_dto.account_info, self.agent_key))
                justification = crypto.decrypt_aes_v2(approval_dto.justification, self.agent_key) if approval_dto.justification else b''
                created = datetime.datetime.fromtimestamp(approval_dto.created / 1000)
                approval = admin_types.PedmApproval(
                    approval_uid=approval_dto.approval_uid, approval_type=approval_dto.approval_type,
                    agent_uid=approval_dto.agent_uid, account_info=account_info,
                    application_info=application_info, justification=justification.decode('utf-8'),
                    expire_in=approval_dto.expire_in, created=created
                )
                approvals.append(approval)
            except Exception as e:
                self.logger.warning('Approval "%s" load error: %s', approval_dto.approval_uid, e)
        if len(approvals) > 0:
            self._approvals.put_entities(approvals)

    def sync_down(self, *, reload: bool = False) -> None:
        if reload:
            self.storage.reset()
            self._populate_data = True

        task = RebuildTask(self._populate_data)
        setting = self.storage.settings.get_entity('PEDM_SYNC_TOKEN')
        if setting is None:
            setting = admin_storage.PedmAdminSettings(key='PEDM_SYNC_TOKEN', value='')
        token: bytes = b''
        if isinstance(setting.value, str) and len(setting.value) > 0:
            token = utils.base64_url_decode(setting.value)

        deployments: List[ admin_storage.PedmStorageDeployment] = []
        delete_deployments: List[str] = []
        policies: List[admin_storage.PedmStoragePolicy] = []
        delete_policies: List[str] = []
        agents: List[admin_storage.PedmStorageAgent] = []
        delete_agents: List[str] = []
        collections: List[admin_storage.PedmStorageCollection] = []
        delete_collections: List[str] = []
        collection_links: List[admin_storage.PedmStorageCollectionLink] = []
        delete_collection_links: List[Tuple[str, str]] = []
        approvals: List[admin_storage.PedmStorageApproval] = []
        delete_approvals: List[str] = []
        approval_status: List[admin_storage.PedmStorageApprovalStatus] = []

        sync_rq = pedm_pb2.GetPedmDataRequest()
        done = False
        while not done:
            sync_rq.continuationToken = token
            sync_rs = api.execute_router(self.params,'pedm/sync_pedm_data', sync_rq, rs_type=pedm_pb2.GetPedmDataResponse)
            assert sync_rs is not None
            if sync_rs.resetCache:
                self.storage.reset()

            token = sync_rs.continuationToken
            done = not sync_rs.hasMore

            for ra in sync_rs.removedAgents:
                agent_uid = utils.base64_url_encode(ra)
                delete_agents.append(agent_uid)

            for rp in sync_rs.removedPolicies:
                policy_uid = utils.base64_url_encode(rp)
                delete_policies.append(policy_uid)

            for rd in sync_rs.removedDeployments:
                deployment_uid = utils.base64_url_encode(rd)
                delete_deployments.append(deployment_uid)

            for rc in sync_rs.removedCollection:
                collection_uid = utils.base64_url_encode(rc)
                delete_collections.append(collection_uid)

            for rcl in sync_rs.removedCollectionLink:
                collection_uid = utils.base64_url_encode(rcl.collectionUid)
                link_uid = utils.base64_url_encode(rcl.linkUid)
                delete_collection_links.append((collection_uid, link_uid))

            for ra in sync_rs.removedApprovals:
                approval_uid = utils.base64_url_encode(ra)
                delete_approvals.append(approval_uid)

            for deployment in sync_rs.deployments:
                dep_id = utils.base64_url_encode(deployment.deploymentUid)
                psd = admin_storage.PedmStorageDeployment(
                    deployment_uid=dep_id, public_key=deployment.ecPublicKey, encrypted_key=deployment.aesKey,
                    disabled=deployment.disabled, data=deployment.encryptedData, created=deployment.created,
                    last_updated=deployment.modified)
                deployments.append(psd)

            for policy in sync_rs.policies:
                policy_uid = utils.base64_url_encode(policy.policyUid)
                policies.append(admin_storage.PedmStoragePolicy(
                    policy_uid=policy_uid, data=policy.encryptedData, admin_data=policy.plainData,
                    key=policy.encryptedKey, disabled=policy.disabled,
                    created=policy.created, updated=policy.modified))

            for agent in sync_rs.agents:
                agent_uid = utils.base64_url_encode(agent.agentUid)
                agents.append(admin_storage.PedmStorageAgent(
                    agent_uid=agent_uid, machine_id=agent.machineId, public_key=agent.ecPublicKey,
                    deployment_uid=utils.base64_url_encode(agent.deploymentUid), data=agent.encryptedData,
                    disabled=agent.disabled, created=agent.created, modified=agent.modified))

            for collection in sync_rs.collections:
                collection_uid = utils.base64_url_encode(collection.collectionUid)
                collections.append(admin_storage.PedmStorageCollection(
                    collection_uid=collection_uid, data=collection.encryptedData,
                    collection_type=collection.collectionType, created=collection.created))

            for collection_link in sync_rs.collectionLink:
                collection_uid = utils.base64_url_encode(collection_link.collectionUid)
                link_uid = utils.base64_url_encode(collection_link.linkUid)
                collection_links.append(admin_storage.PedmStorageCollectionLink(
                    collection_uid=collection_uid, link_uid=link_uid, link_type=collection_link.linkType))

            for approval in sync_rs.approvals:
                approval_uid = utils.base64_url_encode(approval.approvalUid)
                agent_uid = utils.base64_url_encode(approval.agentUid)

                approvals.append(admin_storage.PedmStorageApproval(
                    approval_uid=approval_uid, approval_type=approval.approvalType, agent_uid=agent_uid,
                    account_info=approval.accountInfo, application_info=approval.applicationInfo,
                    justification=approval.justification, expire_in=approval.expireIn, created=approval.created))

            for status in sync_rs.approvalStatus:
                approval_uid = utils.base64_url_encode(status.approvalUid)
                approval_status.append(admin_storage.PedmStorageApprovalStatus(
                    approval_uid=approval_uid, approval_status=status.approvalStatus,
                    enterprise_user_id=status.enterpriseUserId, modified=status.modified))

        setting.value = utils.base64_url_encode(token)
        self.storage.settings.put_entities([setting])

        if len(delete_deployments) > 0:
            uids = list(delete_deployments)
            self.storage.deployments.delete_uids(uids)
        if len(delete_policies) > 0:
            uids = list(delete_policies)
            task.add_policies(uids)
            self.storage.policies.delete_uids(uids)
        if len(delete_agents) > 0:
            task.add_agents(delete_agents)
            delete_approvals.extend(
                [x.approval_uid for x in self.storage.approvals.get_all_entities() if x.agent_uid in delete_agents])
            self.storage.collection_links.delete_links_for_objects(delete_agents)
            self.storage.agents.delete_uids(delete_agents)
        if len(delete_collections) > 0:
            task.add_collections(delete_collections)
            self.storage.collection_links.delete_links_for_subjects(delete_collections)
            self.storage.collection_links.delete_links_for_objects(delete_collections)
            self.storage.collections.delete_uids(delete_collections)
        if len(delete_collection_links) > 0:
            self.storage.collection_links.delete_links(delete_collection_links)
        if len(delete_approvals) > 0:
            task.add_approvals(delete_approvals)
            self.storage.approvals.delete_uids(delete_approvals)
            self.storage.approval_status.delete_uids(delete_approvals)

        if len(deployments) > 0:
            self.storage.deployments.put_entities(deployments)
        if len(policies) > 0:
            task.add_policies((x.policy_uid for x in policies))
            self.storage.policies.put_entities(policies)
        if len(agents) > 0:
            task.add_agents((x.agent_uid for x in agents))
            self.storage.agents.put_entities(agents)
        if len(collections) > 0:
            task.add_collections((x.collection_uid for x in collections))
            self.storage.collections.put_entities(collections)
        if len(collection_links) > 0:
            self.storage.collection_links.put_links(collection_links)
        if len(approvals) > 0:
            task.add_approvals((x.approval_uid for x in approvals))
            self.storage.approvals.put_entities(approvals)
        if len(approval_status) > 0:
            task.add_approvals((x.approval_uid for x in approval_status))
            self.storage.approval_status.put_entities(approval_status)

        self.build_data(task)
        self._need_sync = False
        self._populate_data = False

    def assign_policy_collections(
            self, policies: List[bytes], collections: List[bytes]
    ) -> admin_types.ModifyStatus:
        rq = pedm_pb2.SetPolicyCollectionRequest()
        for policy_uid in policies:
            rq_link = pedm_pb2.PolicyLink()
            rq_link.policyUid = policy_uid
            rq_link.collectionUid.extend(collections)
            rq.setCollection.append(rq_link)

        status_rs = api.execute_router(self.params, rq, 'pedm/set_policy_collections', rs_type=pedm_pb2.PedmStatusResponse)
        self._need_sync = True
        assert status_rs is not None
        return admin_types.ModifyStatus.from_proto(status_rs)

    def modify_policies(self, *,
                        add_policies: Optional[Iterable[admin_types.PedmPolicy]] = None,
                        update_policies: Optional[Iterable[admin_types.PedmUpdatePolicy]] = None,
                        remove_policies: Optional[Iterable[str]] = None) -> admin_types.ModifyStatus:
        rq = pedm_pb2.PolicyRequest()
        if add_policies is not None:
            for policy in add_policies:
                pa = pedm_pb2.PolicyAdd()
                policy_uid = policy.policy_uid or utils.generate_uid()
                pa.policyUid = utils.base64_url_decode(policy_uid)
                policy_key = policy.policy_key or utils.generate_aes_key()
                pa.encryptedKey = crypto.encrypt_aes_v2(policy_key, self.agent_key)

                admin_data = json.dumps(policy.admin_data).encode('utf-8')
                pa.plainData = admin_data

                policy_data = json.dumps(policy.data).encode('utf-8')
                encrypted_data = crypto.encrypt_aes_v2(policy_data, policy_key)
                pa.encryptedData = encrypted_data

                pa.disabled = policy.disabled
                rq.addPolicy.append(pa)
        if update_policies is not None:
            for policy_update in update_policies:
                policy_uid = policy_update.policy_uid
                existing_policy = self.policies.get_entity(policy_uid)
                if existing_policy is None:
                    raise Exception(f'Update: Policy {policy_uid} not found')
                pu = pedm_pb2.PolicyUpdate()
                pu.policyUid = utils.base64_url_decode(existing_policy.policy_uid)
                if isinstance(policy_update.admin_data, dict):
                    pass
                if isinstance(policy_update.data, dict):
                    json_policy = json.dumps(policy_update.data).encode('utf-8')
                    encrypted_data = crypto.encrypt_aes_v2(json_policy, existing_policy.policy_key)
                    pu.encryptedData = encrypted_data
                if isinstance(policy_update.disabled, bool):
                    pu.disabled = folder_pb2.BOOLEAN_TRUE if policy_update.disabled else folder_pb2.BOOLEAN_FALSE
                rq.updatePolicy.append(pu)
        if remove_policies is not None:
            rq.removePolicy.extend((utils.base64_url_decode(x) for x in remove_policies))

        status_rs = api.execute_router(self.params, 'pedm/modify_policy', rq, rs_type=pedm_pb2.PedmStatusResponse)
        self._need_sync = True
        assert status_rs is not None
        return admin_types.ModifyStatus.from_proto(status_rs)


    @staticmethod
    def load_deployment(s_dep: admin_storage.PedmStorageDeployment, tree_key: bytes) -> admin_types.PedmDeployment:
        deployment_key = crypto.decrypt_aes_v2(s_dep.encrypted_key, tree_key)
        decrypted_data = crypto.decrypt_aes_v2(s_dep.data, deployment_key)
        data = pedm_pb2.DeploymentData()
        data.ParseFromString(decrypted_data)
        name = data.name
        d_private_key = data.ecPrivateKey
        created = datetime.datetime.fromtimestamp(s_dep.created / 1000)
        updated = datetime.datetime.fromtimestamp(s_dep.last_updated / 1000)
        return admin_types.PedmDeployment(
            deployment_uid=s_dep.deployment_uid, name=name, deployment_key=deployment_key, disabled=s_dep.disabled,
            created=created, updated=updated, public_key=s_dep.public_key, private_key=d_private_key)

    def modify_deployments(self, *,
                        add_deployments: Optional[Iterable[admin_types.AddDeployment]] = None,
                        update_deployments: Optional[Iterable[admin_types.UpdateDeployment]] = None,
                        remove_deployments: Optional[Iterable[str]] = None) -> admin_types.ModifyStatus:
        enterprise = self.params.enterprise
        tree_key = enterprise['unencrypted_tree_key']

        mrq = pedm_pb2.ModifyDeploymentRequest()
        if add_deployments is not None:
            for add_deployment in add_deployments:
                deployment_uid = utils.generate_uid()
                deployment_key = utils.generate_aes_key()
                priv_key, pub_key = crypto.generate_ec_key()
                d_public_key = crypto.unload_ec_public_key(pub_key)
                d_private_key = crypto.unload_ec_private_key(priv_key)

                a_rq = pedm_pb2.DeploymentCreateRequest()
                a_rq.deploymentUid = utils.base64_url_decode(deployment_uid)
                a_rq.aesKey = crypto.encrypt_aes_v2(deployment_key, tree_key)
                a_rq.ecPublicKey = d_public_key
                if isinstance(add_deployment.spiffe_cert, bytes) and len(add_deployment.spiffe_cert) > 0:
                    a_rq.spiffeCertificate = add_deployment.spiffe_cert
                data = pedm_pb2.DeploymentData()
                data.ecPrivateKey = d_private_key
                if add_deployment.name:
                    data.name = add_deployment.name
                data_bytes = data.SerializeToString()
                a_rq.encryptedData = crypto.encrypt_aes_v2(data_bytes, deployment_key)
                agent_data = json.dumps(add_deployment.agent_info.to_dict()).encode('utf-8')
                a_rq.agentData = crypto.encrypt_ec(agent_data, pub_key)
                mrq.addDeployment.append(a_rq)

        if update_deployments is not None:
            for ud in update_deployments:
                deployment_uid = ud.deployment_uid
                s_dep = self.storage.deployments.get_entity(deployment_uid)
                if not s_dep:
                    raise Exception(f'Update Deployment: "{deployment_uid}" not found')
                dep = self.load_deployment(s_dep, tree_key)

                u_rq = pedm_pb2.DeploymentUpdateRequest()
                u_rq.deploymentUid = utils.base64_url_decode(ud.deployment_uid)
                if ud.disabled is None:
                    u_rq.disabled = folder_pb2.BOOLEAN_NO_CHANGE
                else:
                    u_rq.disabled = folder_pb2.BOOLEAN_TRUE if ud.disabled else folder_pb2.BOOLEAN_FALSE
                if ud.name:
                    data = pedm_pb2.DeploymentData()
                    data.ecPrivateKey =dep.private_key
                    data.name = ud.name
                    data_bytes = data.SerializeToString()
                    u_rq.encryptedData = crypto.encrypt_aes_v2(data_bytes, dep.deployment_key)
                if ud.spiffe_cert is not None:
                    u_rq.spiffeCertificate = ud.spiffe_cert

                mrq.updateDeployment.append(u_rq)

        if remove_deployments is not None:
            for deployment_uid in remove_deployments:
                s_dep = self.storage.deployments.get_entity(deployment_uid)
                if not s_dep:
                    raise Exception(f'Delete Deployment: "{deployment_uid}" not found')
                mrq.removeDeployment.append(utils.base64_url_decode(deployment_uid))

        status_rs = api.execute_router(self.params,'pedm/modify_deployment', request=mrq, rs_type=pedm_pb2.PedmStatusResponse)
        assert status_rs is not None
        self._need_sync = True
        return admin_types.ModifyStatus.from_proto(status_rs)

    def modify_collections(self, *,
                        add_collections: Optional[Iterable[admin_types.CollectionData]] = None,
                        update_collections: Optional[Iterable[admin_types.CollectionData]] = None,
                        remove_collections: Optional[Iterable[str]] = None) -> admin_types.ModifyStatus:
        to_add: List[pedm_pb2.CollectionValue] = []
        to_update: List[pedm_pb2.CollectionValue] = []
        for colls in (add_collections, update_collections):
            if colls is not None:
                for coll in colls:
                    cv = pedm_pb2.CollectionValue()
                    cv.collectionUid = utils.base64_url_decode(coll.collection_uid)
                    cv.collectionType = coll.collection_type
                    cv.encryptedData = crypto.encrypt_aes_v2(coll.collection_data.encode(), self.agent_key)
                    if colls is add_collections:
                        to_add.append(cv)
                    elif colls is update_collections:
                        to_update.append(cv)
        to_remove: List[bytes] = []
        if remove_collections is not None:
            for collection_uid in remove_collections:
                to_remove.append(utils.base64_url_decode(collection_uid))

        status = admin_types.ModifyStatus(add=[], update=[], remove=[])
        while len(to_add) > 0 or len(to_update) > 0 or len(to_remove) > 0:
            crq = pedm_pb2.CollectionRequest()
            if len(to_add) > 0:
                add_chunk = to_add[:500]
                to_add = to_add[500:]
                crq.addCollection.extend(add_chunk)

            if len(to_update) > 0:
                update_chunk = to_update[:500]
                to_update = to_update[500:]
                crq.updateCollection.extend(update_chunk)

            if len(to_remove) > 0:
                remove_chunk = to_remove[:500]
                to_remove = to_remove[500:]
                crq.removeCollection.extend(remove_chunk)
            status_rs = api.execute_router(self.params, 'pedm/modify_collection', request=crq,
                                           rs_type=pedm_pb2.PedmStatusResponse)
            assert status_rs is not None
            status.merge(admin_types.ModifyStatus.from_proto(status_rs))
        self._need_sync = True
        return status

    def get_collection_links(self, *, links: Iterable[admin_types.CollectionLink]) -> Iterable[admin_types.CollectionLinkData]:
        link_rq = pedm_pb2.GetCollectionLinkRequest()
        for l in links:
            cl = pedm_pb2.CollectionLink()
            cl.collectionUid = utils.base64_url_decode(l.collection_uid)
            cl.linkType = cast(pedm_pb2.CollectionLinkType, l.link_type)
            cl.linkUid = utils.base64_url_decode(l.link_uid)
            link_rq.collectionLink.append(cl)
        link_rs = api.execute_router(self.params, 'pedm/get_collection_links', request=link_rq,
                                     rs_type=pedm_pb2.GetCollectionLinkResponse)
        assert link_rs is not None
        for ld in link_rs.collectionLinkData:
            collection_link = admin_types.CollectionLink(
                collection_uid=utils.base64_url_encode(ld.collectionUid), link_type=ld.linkType,
                link_uid=utils.base64_url_encode(ld.linkUid))
            yield admin_types.CollectionLinkData(collection_link=collection_link, link_data=ld.linkData)

    def set_collection_links(
            self, *, set_links: Optional[Iterable[admin_types.CollectionLink]] = None,
            unset_links: Optional[Iterable[admin_types.CollectionLink]] = None
    ) -> admin_types.ModifyStatus:
        clrq = pedm_pb2.SetCollectionLinkRequest()
        if set_links is not None:
            for coll in set_links:
                cln = pedm_pb2.CollectionLinkData()
                cln.collectionUid = utils.base64_url_decode(coll.collection_uid)
                cln.linkUid = utils.base64_url_decode(coll.link_uid)
                cln.linkType = coll.link_type     # type: ignore
                clrq.addCollection.append(cln)

        if unset_links is not None:
            for coll in unset_links:
                cl = pedm_pb2.CollectionLink()
                cl.collectionUid = utils.base64_url_decode(coll.collection_uid)
                cl.linkUid = utils.base64_url_decode(coll.link_uid)
                cl.linkType = coll.link_type      # type: ignore
                clrq.removeCollection.append(cl)

        status_rs = api.execute_router(self.params, 'pedm/set_collection_links', request=clrq,
                                       rs_type=pedm_pb2.PedmStatusResponse)
        assert status_rs is not None
        self._need_sync = True
        return admin_types.ModifyStatus.from_proto(status_rs)

    def modify_agents(self, *,
                        update_agents: Optional[Iterable[admin_types.UpdateAgent]] = None,
                        remove_agents: Optional[Iterable[str]] = None) -> admin_types.ModifyStatus:
        rq = pedm_pb2.ModifyAgentRequest()
        if update_agents is not None:
            for ua in update_agents:
                agent_uid = ua.agent_uid
                existing_agent = self.agents.get_entity(agent_uid)
                if existing_agent is None:
                    raise Exception(f'Update: Policy {agent_uid} not found')
                au = pedm_pb2.AgentUpdate()
                au.agentUid = utils.base64_url_decode(agent_uid)
                if isinstance(ua.deployment_uid, str):
                    au.deploymentUid = utils.base64_url_decode(ua.deployment_uid)
                if isinstance(ua.disabled, bool):
                    au.disabled = folder_pb2.BOOLEAN_TRUE if ua.disabled else folder_pb2.BOOLEAN_FALSE

                rq.updateAgent.append(au)
        if remove_agents is not None:
            rq.removeAgent.extend((utils.base64_url_decode(x) for x in remove_agents))

        status_rs = api.execute_router(self.params,'pedm/modify_agent', rq, rs_type=pedm_pb2.PedmStatusResponse)
        self._need_sync = True
        assert status_rs is not None
        return admin_types.ModifyStatus.from_proto(status_rs)

    def modify_approvals(self, *,
                        to_approve: Optional[List[bytes]] = None,
                        to_deny: Optional[List[bytes]] = None,
                        to_remove: Optional[List[bytes]] = None) -> admin_types.ModifyStatus:
        rq = pedm_pb2.ApprovalActionRequest()
        if to_approve:
            rq.approve.extend(to_approve)
        if to_deny:
            rq.deny.extend(to_deny)
        if to_remove:
            rq.remove.extend(to_remove)

        status_rs = api.execute_router(self.params, 'pedm/approval_action', rq, rs_type=pedm_pb2.PedmStatusResponse)
        self._need_sync = True
        assert status_rs is not None
        return admin_types.ModifyStatus.from_proto(status_rs)


def get_pedm_plugin(context: KeeperParams) -> PedmPlugin:
    if context._pedm_plugin is None:
        context._pedm_plugin = PedmPlugin(context)
    if context._pedm_plugin.need_sync:
        context._pedm_plugin.sync_down()
    return context._pedm_plugin



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import csv
import os
import zipfile
from typing import Union, Iterable

from .. import importer
from ... import vault


class DashlaneImporter(importer.BaseFileImporter):
    def do_import(self, filename, **kwargs):
        # type: (str, dict) -> Iterable[Union[importer.Record, importer.SharedFolder, importer.File]]
        
        # Handle zip file containing multiple CSV files
        if filename.endswith('.zip'):
            with zipfile.ZipFile(filename, 'r') as zf:
                # Extract files to temporary directory
                temp_dir = kwargs.get('tmpdir') or os.path.dirname(filename)
                zf.extractall(temp_dir)
                
                # Check for files in root or in dashlane-credential-export subdirectory
                base_dirs = ['', 'dashlane-credential-export']
                
                # Process each CSV file
                for csv_file in ['credentials.csv', 'securenotes.csv', 'payments.csv', 'ids.csv', 'personalinfo.csv']:
                    for base_dir in base_dirs:
                        csv_path = os.path.join(temp_dir, base_dir, csv_file)
                        if os.path.exists(csv_path):
                            yield from self._process_csv_file(csv_path, csv_file)
                            break  # Found the file, no need to check other directories
        else:
            # Handle individual CSV file
            yield from self._process_csv_file(filename, os.path.basename(filename))

    def _process_csv_file(self, filename, file_type):
        with open(filename, 'r', encoding='utf-8') as csvfile:
            reader = csv.DictReader(csvfile)
            
            if file_type == 'credentials.csv':
                yield from self._process_credentials(reader)
            elif file_type == 'securenotes.csv':
                yield from self._process_secure_notes(reader)
            elif file_type == 'payments.csv':
                yield from self._process_payments(reader)
            elif file_type == 'ids.csv':
                yield from self._process_ids(reader)
            elif file_type == 'personalinfo.csv':
                yield from self._process_personal_info(reader)

    def _process_credentials(self, reader):
        for row in reader:
            record = importer.Record()
            record.type = 'login'
            record.title = row.get('title', '').strip()
            record.login = row.get('username', '').strip()
            record.password = row.get('password', '').strip()
            record.login_url = row.get('url', '').strip()
            record.notes = row.get('note', '').strip()
            
            # Handle 2FA
            otp_url = row.get('otpUrl', '').strip()
            if otp_url:
                record.fields.append(importer.RecordField('oneTimeCode', '', otp_url))
            
            # Handle additional usernames
            for i in range(2, 4):  # username2 and username3
                username = row.get(f'username{i}', '').strip()
                if username:
                    record.fields.append(importer.RecordField('text', f'username{i}', username))
            
            yield record

    def _process_secure_notes(self, reader):
        for row in reader:
            record = importer.Record()
            record.type = 'encryptedNotes'
            record.title = row.get('title', '').strip()
            record.notes = row.get('note', '').strip()
            
            # Add category as a custom field
            category = row.get('category', '').strip()
            if category:
                record.fields.append(importer.RecordField('text', 'category', category))
            
            yield record

    def _process_payments(self, reader):
        for row in reader:
            record = importer.Record()
            record_type = row.get('type', '').strip().lower()
            
            if record_type == 'payment_card':
                record.type = 'bankCard'
                card = {
                    'cardNumber': row.get('cc_number', '').strip(),
                    'cardExpirationDate': f"{row.get('expiration_month', '').strip()}/{row.get('expiration_year', '').strip()}",
                    'cardSecurityCode': row.get('code', '').strip()
                }
                record.fields.append(importer.RecordField('paymentCard', '', card))
                record.title = row.get('name', '').strip()
            elif record_type == 'bank_account':
                record.type = 'bankAccount'
                account = {
                    'accountNumber': row.get('account_number', '').strip(),
                    'routingNumber': row.get('routing_number', '').strip()
                }
                record.fields.append(importer.RecordField('bankAccount', '', account))
                record.title = row.get('account_name', '').strip()
            
            # Add common fields
            if row.get('account_holder'):
                record.fields.append(importer.RecordField('text', 'accountHolder', row['account_holder']))
            if row.get('issuing_bank'):
                record.fields.append(importer.RecordField('text', 'issuingBank', row['issuing_bank']))
            if row.get('note'):
                record.notes = row['note']
            
            yield record

    def _process_ids(self, reader):
        for row in reader:
            record = importer.Record()
            record_type = row.get('type', '').strip().lower()
            
            if record_type == 'license':
                record.type = 'driverLicense'
                record.title = f"Driver's License - {row.get('name', '').strip()}"
            elif record_type == 'passport':
                record.type = 'passport'
                record.title = f"Passport - {row.get('name', '').strip()}"
            else:
                # For card, social_security, tax_number - use encryptedNotes since there's no direct match
                record.type = 'encryptedNotes'
                record.title = f"{record_type.title()} - {row.get('name', '').strip()}"
            
            # Add common fields
            if row.get('number'):
                record.fields.append(importer.RecordField('text', 'number', row['number']))
            if row.get('issue_date'):
                record.fields.append(importer.RecordField('date', 'issueDate', row['issue_date']))
            if row.get('expiration_date'):
                record.fields.append(importer.RecordField('date', 'expirationDate', row['expiration_date']))
            if row.get('place_of_issue'):
                record.fields.append(importer.RecordField('text', 'placeOfIssue', row['place_of_issue']))
            if row.get('state'):
                record.fields.append(importer.RecordField('text', 'state', row['state']))
            
            yield record

    def _process_personal_info(self, reader):
        # Group contact information by person (using login as identifier)
        contact_records = {}
        address_records = []
        
        for row in reader:
            record_type = row.get('type', '').strip().lower()
            
            if record_type == 'name':
                # Create or update contact record
                login = row.get('login', '').strip()
                if not login:
                    continue
                    
                if login not in contact_records:
                    contact_records[login] = {
                        'type': 'contact',  # Keeper's contact record type
                        'title': '',
                        'fields': [],
                        'notes': ''
                    }
                
                record = contact_records[login]
                first_name = row.get('first_name', '').strip()
                last_name = row.get('last_name', '').strip()
                record['title'] = f"{first_name} {last_name}".strip() or f"Contact - {login}"
                
                name = {
                    'first': first_name,
                    'middle': row.get('middle_name', ''),
                    'last': last_name
                }
                record['fields'].append(importer.RecordField('name', '', name))
                
                # Add date of birth if available
                dob = row.get('date_of_birth', '').strip()
                if dob:
                    record['fields'].append(importer.RecordField('date', 'birthDate', dob))
                
                # Add place of birth if available
                pob = row.get('place_of_birth', '').strip()
                if pob:
                    record['fields'].append(importer.RecordField('text', 'placeOfBirth', pob))
                
                # Add job title if available
                job = row.get('job_title', '').strip()
                if job:
                    record['fields'].append(importer.RecordField('text', 'jobTitle', job))
                
                # Add URL if available
                url = row.get('url', '').strip()
                if url:
                    record['fields'].append(importer.RecordField('url', '', url))
            
            elif record_type == 'email':
                # Add email to contact record
                login = row.get('login', '').strip()
                if login in contact_records:
                    email = row.get('email', '').strip()
                    if email:
                        contact_records[login]['fields'].append(importer.RecordField('email', '', email))
            
            elif record_type == 'phone':
                # Add phone to contact record
                login = row.get('login', '').strip()
                if login in contact_records:
                    phone = row.get('phone_number', '').strip()
                    if phone:
                        contact_records[login]['fields'].append(importer.RecordField('phone', '', phone))
            
            elif record_type == 'address':
                # Create a separate address record
                address_record = {
                    'type': 'address',  # Keeper's address record type
                    'title': '',
                    'fields': [],
                    'notes': ''
                }
                
                # Set title
                item_name = row.get('item_name', '').strip()
                address_record['title'] = f"Address - {item_name}" if item_name else "Address"
                
                # Add address fields
                address = {
                    'street1': row.get('address', ''),
                    'street2': row.get('address_apartment', ''),
                    'city': row.get('city', ''),
                    'state': row.get('state', ''),
                    'zip': row.get('zip', ''),
                    'country': row.get('country', '')
                }
                address_record['fields'].append(importer.RecordField('address', '', address))
                
                # Add recipient if available
                recipient = row.get('address_recipient', '').strip()
                if recipient:
                    address_record['fields'].append(importer.RecordField('text', 'recipient', recipient))
                
                address_records.append(address_record)
        
        # Yield contact records
        for login, data in contact_records.items():
            record = importer.Record()
            record.type = data['type']
            record.title = data['title']
            record.fields = data['fields']
            record.notes = data['notes']
            yield record
        
        # Yield address records
        for data in address_records:
            record = importer.Record()
            record.type = data['type']
            record.title = data['title']
            record.fields = data['fields']
            record.notes = data['notes']
            yield record

    def extension(self):
        return 'zip' 


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example code to onboard a user
# 1. Approve teams for a new user
# 2. Check for "enterprise_push.json" file
#
# Usage:
#    python3 user_onboarding.py

import argparse
import getpass
import logging
import os
import sys
import time
from typing import Set

from keepercommander import api
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands.enterprise import TeamApproveCommand
from keepercommander.params import KeeperParams
from keepercommander.commands.enterprise_push import EnterprisePushCommand


parser = argparse.ArgumentParser(description='Onboard users')
parser.add_argument('--debug', action='store_true', help='Enables debug logging')
opts, flags = parser.parse_known_args(sys.argv[1:])
logging.basicConfig(level=logging.DEBUG if opts.debug is True else logging.INFO, format='%(message)s')


if os.path.isfile('enterprise_push.json'):
    logging.debug('Enterprise push file detected.')


def approve_teams(params):
    command = TeamApproveCommand()
    command.execute(params, team=True, user=True)


def push_records_to_user(params, enterprise_user_id):    # type: (KeeperParams, int) -> None
    username = next((x['username'] for x in params.enterprise['users'] if x['enterprise_user_id'] == enterprise_user_id), None)
    if username:
        command = EnterprisePushCommand()
        command.execute(params, user=[username], file='enterprise_push.json')
    else:
        logging.debug(f'Cannot resolve username for {enterprise_user_id}')


my_params = get_params_from_config(os.path.join(os.path.dirname(__file__), 'config.json'))
while not my_params.user:
    my_params.user = getpass.getpass(prompt='User(Email): ', stream=None)

api.login(my_params)
if not my_params.session_token:
    exit(1)

if not my_params.enterprise:
    print('Not an enterprise administrator')
    exit(1)

active_users = {x['enterprise_user_id'] for x in my_params.enterprise['users']
                if x['status'] == 'active' and x.get('lock') == 0}    # type: Set[int]
new_users = set()    # type: Set[int]

# detect new users by checking queued users
if 'teams' in my_params.enterprise:
    teams = {x['team_uid'] for x in my_params.enterprise['teams']}
    if 'queued_team_users' in my_params.enterprise:
        for qu in my_params.enterprise['queued_team_users']:
            if qu['team_uid'] in teams:
                new_users.update((x for x in qu['users'] if x in active_users))


while True:
    if len(new_users) > 0:
        print(f'Provisioning {len(new_users)} users')
        approve_teams(my_params)
        if os.path.isfile('enterprise_push.json'):
            for user_id in new_users:
                try:
                    push_records_to_user(my_params, user_id)
                except Exception as e:
                    print(f'Cannot push records to user ID {user_id}: {e}')
        new_users.clear()

    logging.debug('Sleeping for 5 minutes.')
    time.sleep(5 * 60)

    api.query_enterprise(my_params)
    au = {x['enterprise_user_id'] for x in my_params.enterprise['users']
          if x['status'] == 'active' and x.get('lock') == 0}
    new_users.update(au.difference(active_users))
    active_users = au



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: client.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63lient.proto\x12\x06Tokens\"u\n\x18\x42reachWatchUpdateRequest\x12\x42\n\x18\x62reachWatchRecordRequest\x18\x01 \x03(\x0b\x32 .Tokens.BreachWatchRecordRequest\x12\x15\n\rencryptedData\x18\x02 \x01(\x0c\"\x9c\x01\n\x18\x42reachWatchRecordRequest\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x15\n\rencryptedData\x18\x02 \x01(\x0c\x12\x38\n\x13\x62reachWatchInfoType\x18\x03 \x01(\x0e\x32\x1b.Tokens.BreachWatchInfoType\x12\x1c\n\x14updateUserWhoScanned\x18\x04 \x01(\x08\"\x81\x01\n\x0f\x42reachWatchData\x12%\n\tpasswords\x18\x01 \x03(\x0b\x32\x12.Tokens.BWPassword\x12\"\n\x06\x65mails\x18\x02 \x03(\x0b\x32\x12.Tokens.BWPassword\x12#\n\x07\x64omains\x18\x03 \x03(\x0b\x32\x12.Tokens.BWPassword\"]\n\nBWPassword\x12\r\n\x05value\x18\x01 \x01(\t\x12\x10\n\x08resolved\x18\x02 \x01(\x04\x12 \n\x06status\x18\x03 \x01(\x0e\x32\x10.Tokens.BWStatus\x12\x0c\n\x04\x65uid\x18\x04 \x01(\x0c*9\n\x13\x42reachWatchInfoType\x12\n\n\x06RECORD\x10\x00\x12\x16\n\x12\x41LTERNATE_PASSWORD\x10\x01*E\n\x08\x42WStatus\x12\x08\n\x04GOOD\x10\x00\x12\x0b\n\x07\x43HANGED\x10\x01\x12\x08\n\x04WEAK\x10\x02\x12\x0c\n\x08\x42REACHED\x10\x03\x12\n\n\x06IGNORE\x10\x04\x42\"\n\x18\x63om.keepersecurity.protoB\x06\x43lientb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'client_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\006Client'
  _globals['_BREACHWATCHINFOTYPE']._serialized_start=529
  _globals['_BREACHWATCHINFOTYPE']._serialized_end=586
  _globals['_BWSTATUS']._serialized_start=588
  _globals['_BWSTATUS']._serialized_end=657
  _globals['_BREACHWATCHUPDATEREQUEST']._serialized_start=24
  _globals['_BREACHWATCHUPDATEREQUEST']._serialized_end=141
  _globals['_BREACHWATCHRECORDREQUEST']._serialized_start=144
  _globals['_BREACHWATCHRECORDREQUEST']._serialized_end=300
  _globals['_BREACHWATCHDATA']._serialized_start=303
  _globals['_BREACHWATCHDATA']._serialized_end=432
  _globals['_BWPASSWORD']._serialized_start=434
  _globals['_BWPASSWORD']._serialized_end=527
# @@protoc_insertion_point(module_scope)



# __  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json

from typing import Union, Iterable

from .. import importer
from ...record_types import FieldTypes


class BitwardenImporter(importer.BaseFileImporter):
    def do_import(self, filename, **kwargs):
        # type: (importer.BaseImporter, str, dict) -> Iterable[Union[importer.Record]]

        with open(filename, 'r', encoding='utf-8') as bw_file:
            bw_import = json.load(bw_file)
        if 'encrypted' not in bw_import or 'items' not in bw_import:
            raise Exception(f'File {filename}: Not a Bitwarden JSON export file')

        if bw_import['encrypted'] is True:
            raise Exception(f'File {filename}: Encrypted Bitwarden JSON export not supported.')

        folders = {}
        for f in bw_import.get('folders', []):
            if 'id' in f and 'name' in f:
                folders[f['id']] = f['name']

        for i in bw_import.get('items') or []:
            record = importer.Record()
            item_type = i.get('type', 0)
            if item_type == 1:
                record.type = 'login'
            elif item_type == 2:
                record.type = 'encryptedNotes'
            elif item_type == 3:
                record.type = 'bankCard'
            elif item_type == 4:
                record.type = 'address'
            else:
                continue

            record.title = i.get('name') or ''
            note = i.get('notes') or ''
            if note:
                if record.type == 'encryptedNotes':
                    record.fields.append(importer.RecordField('note', '', note))
                else:
                    record.notes = note

            folder_id = i.get('folderId') or ''
            if folder_id in folders:
                f = importer.Folder()
                f.path = folders[folder_id]
                record.folders = [f]
            fields = i.get('fields')
            if isinstance(fields, list):
                for field in fields:
                    field_type = field.get('type', -1)
                    field_name = field.get('name')
                    if not field_name:
                        continue
                    field_value = field.get('value') or ''
                    if field_type == 0:
                        record.fields.append(importer.RecordField('text', field_name, field_value))
                    elif field_type == 1:
                        record.fields.append(importer.RecordField('secret', field_name, field_value))
                    elif field_type == 2:
                        field_value = field_value == 'true'
                        record.fields.append(importer.RecordField('checkbox', field_name, field_value))

            login = i.get('login')
            if isinstance(login, dict):
                record.login = login.get('username') or ''
                record.password = login.get('password') or ''
                totp = login.get('totp')
                if totp:
                    if not totp.startswith('otpauth://'):
                        totp = f'otpauth://totp/?secret={totp}'
                    record.fields.append(importer.RecordField('oneTimeCode', '', totp))
                uris = login.get('uris')
                if isinstance(uris, list):
                    for u in uris:
                        if isinstance(u, dict):
                            uri = u.get('uri')
                            if uri:
                                if record.login_url:
                                    record.fields.append(importer.RecordField('url', '', uri))
                                else:
                                    record.login_url = uri

            identity = i.get('identity')
            if isinstance(identity, dict):
                f_name = identity.get('firstName') or ''
                l_name = identity.get('lastName') or ''
                if f_name or l_name:
                    name = FieldTypes['name'].value.copy()
                    name['first'] = f_name
                    name['last'] = l_name
                    name['middle'] = identity.get('middleName') or ''
                    record.fields.append(importer.RecordField('name', '', name))
                address1 = identity.get('address1') or ''
                city = identity.get('city') or ''
                state = identity.get('state') or ''
                postal_code = identity.get('postalCode') or ''
                if address1 or city or state or postal_code:
                    address = FieldTypes['address'].value.copy()
                    address['street1'] = address1
                    address['street2'] = identity.get('address2') or ''
                    address['city'] = city
                    address['state'] = state
                    address['zip'] = postal_code
                    address['country'] = identity.get('country') or ''
                    record.fields.append(importer.RecordField('address', '', address))
                company = identity.get('company') or ''
                if company:
                    record.fields.append(importer.RecordField('text', 'company', company))
                email = identity.get('email') or ''
                if email:
                    record.fields.append(importer.RecordField('email', '', email))
                phone = identity.get('phone') or ''
                if phone:
                    phone_number = FieldTypes['phone'].value.copy()
                    phone_number['number'] = phone
                    record.fields.append(importer.RecordField('phone', '', phone_number))
                ssn = identity.get('ssn') or ''
                if ssn:
                    record.fields.append(importer.RecordField('accountNumber', 'identityNumber', ssn))
                username = identity.get('username') or ''
                if username:
                    if record.login:
                        record.fields.append(importer.RecordField('login', '', username))
                    else:
                        record.login = username
                passport_number = identity.get('passportNumber') or ''
                if passport_number:
                    record.fields.append(importer.RecordField('accountNumber', 'passportNumber', passport_number))
                license_number = identity.get('licenseNumber') or ''
                if license_number:
                    record.fields.append(importer.RecordField('accountNumber', 'licenseNumber', license_number))

            card = i.get('card')
            if isinstance(card, dict):
                cardholder = card.get('cardholderName') or ''
                if cardholder:
                    record.fields.append(importer.RecordField('text', 'cardholderName', cardholder))

                number = card.get('number') or ''
                code = card.get('code') or ''
                if number or code:
                    payment_card = FieldTypes['paymentCard'].value.copy()
                    payment_card['cardNumber'] = number
                    payment_card['cardSecurityCode'] = code
                    month = card.get('expMonth') or ''
                    year = card.get('expYear') or ''
                    if month and year:
                        if len(month) == 1:
                            month = '0' + month
                    payment_card['cardExpirationDate'] = f'{month}/{year}'
                    record.fields.append(importer.RecordField('paymentCard', '', payment_card))

            yield record

    def extension(self):
        return 'json'




import json
import logging

from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError

from ..loader import SecureStorageBase, SecureStorageException

logging.getLogger('botocore').setLevel(logging.WARNING)


class SecretsManagerStorage(SecureStorageBase):
    @staticmethod
    def get_endpoint(url):    # type: (str) -> Tuple[str, str]
        res = urlparse(url)
        if res.scheme != 'aws-sm':
            raise ValueError(f'Invalid URL scheme. "aws-sm" is expected.')

        secret_key = res.path
        if secret_key.startswith('/'):
            secret_key = secret_key[1:]
        region_name = res.netloc
        return region_name, secret_key

    def load_configuration(self, url, storage_value=None):
        region_name, secret_name = SecretsManagerStorage.get_endpoint(url)
        try:
            session = boto3.session.Session()
            client = session.client(service_name='secretsmanager', region_name=region_name)
            value = client.get_secret_value(SecretId=secret_name)
            if 'SecretString' in value:
                return json.loads(value['SecretString'])
            return {}
        except ClientError as e:
            raise SecureStorageException(str(e))

    def store_configuration(self, url, configuration):
        region_name, secret_name = SecretsManagerStorage.get_endpoint(url)
        try:
            session = boto3.session.Session()
            client = session.client(service_name='secretsmanager', region_name=region_name)

            value = json.dumps(configuration)
            client.put_secret_value(SecretId=secret_name, SecretString=value)
        except ClientError as e:
            raise SecureStorageException(str(e))
